import unittest
import numpy as np
import pandas as pd

from lightwood.api.types import ProblemDefinition
from lightwood.api.high_level import json_ai_from_problem, predictor_from_json_ai
from lightwood.helpers.imputers import NumericalImputer, CategoricalImputer


class TestCleaner(unittest.TestCase):
    def test_0_imputers(self):

        df = pd.read_csv('tests/data/hdi.csv')[:100]
        df = df.rename(columns={'GDP ($ per capita)': 'GDP', 'Area (sq. mi.)': 'Area', 'Literacy (%)': 'Literacy'})
        target = 'Development Index'
        df['Infant mortality '] = df['Infant mortality '].apply(lambda x: 'High' if x >= 20 else 'Low')
        jai = json_ai_from_problem(df, ProblemDefinition.from_dict({'target': target, 'time_aim': 10}))

        # define columns to impute
        cat_mode_impute_col = 'Infant mortality '
        num_mean_impute_col = 'Population'
        num_mode_impute_col = 'GDP'
        num_zero_impute_col = 'Pop. Density '
        num_median_impute_col = 'Area'
        cols = [num_mean_impute_col, num_mode_impute_col, num_zero_impute_col, num_median_impute_col,
                cat_mode_impute_col]

        # replace first row values with nans
        for col in cols:
            df[col].iloc[0] = np.nan

        num_mean_target_value = df[num_mean_impute_col].iloc[1:].mean()
        num_mode_target_value = df[num_mode_impute_col].iloc[1:].mode().iloc[0]
        num_median_target_value = df[num_median_impute_col].iloc[1:].median()
        num_zero_target_value = 0.0
        cat_mode_target_value = df[cat_mode_impute_col].iloc[1:].mode().iloc[0]

        jai.imputers = [
            {"module": "NumericalImputer", "args": {"value": "'mean'", "target": f"'{num_mean_impute_col}'"}},
            {"module": "NumericalImputer", "args": {"value": "'mode'", "target": f"'{num_mode_impute_col}'"}},
            {"module": "NumericalImputer", "args": {"value": "'median'", "target": f"'{num_median_impute_col}'"}},
            {"module": "NumericalImputer", "args": {"value": "'zero'", "target": f"'{num_zero_impute_col}'"}},
            {"module": "CategoricalImputer", "args": {"value": "'mode'", "target": f"'{cat_mode_impute_col}'"}},
        ]
        predictor = predictor_from_json_ai(jai)
        cleaned_data = predictor.preprocess(df)

        # check cleaner was assigned imputers and data was imputed correctly
        for _, imputer in predictor.imputers.items():
            print(type(imputer))
            assert isinstance(imputer, NumericalImputer) or isinstance(imputer, CategoricalImputer)

        assert cleaned_data[num_mean_impute_col].iloc[0] == num_mean_target_value
        assert cleaned_data[num_mode_impute_col].iloc[0] == num_mode_target_value
        assert cleaned_data[num_zero_impute_col].iloc[0] == num_zero_target_value
        assert cleaned_data[num_median_impute_col].iloc[0] == num_median_target_value
        assert cleaned_data[cat_mode_impute_col].iloc[0] == cat_mode_target_value



import unittest
import pandas as pd
from sklearn.metrics import r2_score, accuracy_score
from lightwood.api.high_level import code_from_json_ai, json_ai_from_problem, predictor_from_code
from lightwood.api.types import ProblemDefinition


class TestBasic(unittest.TestCase):
    def test_0_mean_ensemble(self):

        df = pd.read_csv('tests/data/concrete_strength.csv')

        target = 'concrete_strength'

        json_ai = json_ai_from_problem(df, ProblemDefinition.from_dict({
            'target': target,
            'time_aim': 80
        }))

        json_ai.model = {
            'module': 'MeanEnsemble',
            "args": {
                'dtype_dict': '$dtype_dict',
                'submodels': json_ai.model['args']['submodels']
            }
        }

        code = code_from_json_ai(json_ai)
        predictor = predictor_from_code(code)
        predictor.learn(df)
        predictions = predictor.predict(df)

        self.assertTrue(r2_score(df[target], predictions['prediction']) > 0.5)

    def test_1_mode_ensemble(self):

        df = pd.read_csv('tests/data/hdi.csv')

        target = 'Development Index'

        json_ai = json_ai_from_problem(df, ProblemDefinition.from_dict({
            'target': target,
            'time_aim': 5
        }))

        json_ai.model = {
            'module': 'ModeEnsemble',
            "args": {
                'dtype_dict': '$dtype_dict',
                "args": "$pred_args",
                "accuracy_functions": "$accuracy_functions",
                'submodels': json_ai.model['args']['submodels']
            }
        }

        code = code_from_json_ai(json_ai)
        predictor = predictor_from_code(code)
        predictor.learn(df)
        predictions = predictor.predict(df)

        self.assertTrue(accuracy_score(df[target].astype(int), predictions['prediction'].astype(int)) > 0.5)

    def test_2_weighted_mean_ensemble(self):
        df = pd.read_csv('tests/data/concrete_strength.csv')

        target = 'concrete_strength'

        json_ai = json_ai_from_problem(df, ProblemDefinition.from_dict({
            'target': target,
            'time_aim': 80
        }))

        json_ai.model = {
            'module': 'WeightedMeanEnsemble',
            "args": {
                'dtype_dict': '$dtype_dict',
                "args": "$pred_args",
                "accuracy_functions": "$accuracy_functions",
                'submodels': json_ai.model['args']['submodels']
            }
        }

        code = code_from_json_ai(json_ai)
        predictor = predictor_from_code(code)
        predictor.learn(df)
        predictions = predictor.predict(df)

        self.assertTrue(r2_score(df[target], predictions['prediction']) > 0.5)



# TODO: _add_implicit_values unit test ensures NO changes for a fully specified file.
from copy import deepcopy
from lightwood.helpers.templating import call, inline_dict, align
from lightwood.api import dtype
from lightwood.api.types import (
    JsonAI,
    TypeInformation,
    StatisticalAnalysis,
    ProblemDefinition,
)
import inspect
from lightwood.helpers.log import log


# For custom modules, we create a module loader with necessary imports below
IMPORT_EXTERNAL_DIRS = """
for import_dir in [os.path.join(os.path.expanduser('~/lightwood_modules'), lightwood_version.replace('.', '_')), os.path.join('/etc/lightwood_modules', lightwood_version.replace('.', '_'))]:
    if os.path.exists(import_dir) and os.access(import_dir, os.R_OK):
        for file_name in list(os.walk(import_dir))[0][2]:
            if file_name[-3:] != '.py':
                continue
            mod_name = file_name[:-3]
            loader = importlib.machinery.SourceFileLoader(mod_name,
                                                          os.path.join(import_dir, file_name))
            module = ModuleType(loader.name)
            loader.exec_module(module)
            sys.modules[mod_name] = module
            exec(f'import {mod_name}')
""" # noqa

IMPORTS = """
import lightwood
from lightwood import __version__ as lightwood_version
from lightwood.analysis import *
from lightwood.api import *
from lightwood.data import *
from lightwood.encoder import *
from lightwood.ensemble import *
from lightwood.helpers.device import *
from lightwood.helpers.general import *
from lightwood.helpers.log import *
from lightwood.helpers.numeric import *
from lightwood.helpers.imputers import *
from lightwood.helpers.parallelism import *
from lightwood.helpers.seed import *
from lightwood.helpers.text import *
from lightwood.helpers.torch import *
from lightwood.mixer import *
import pandas as pd
from typing import Dict, List, Union
import os
from types import ModuleType
import importlib.machinery
import sys
import time
"""


def lookup_encoder(
    col_dtype: str,
    col_name: str,
    is_target: bool,
    problem_defintion: ProblemDefinition,
    is_target_predicting_encoder: bool,
    statistical_analysis: StatisticalAnalysis,
):
    """
    Assign a default encoder for a given column based on its data type, and whether it is a target. Encoders intake raw (but cleaned) data and return an feature representation. This function assigns, per data type, what the featurizer should be. This function runs on each column within the dataset available for model building to assign how it should be featurized.

    Users may override to create a custom encoder to enable their own featurization process. However, in order to generate template JSON-AI, this code runs automatically. Users may edit the generated syntax and use custom approaches while model building.

    For each encoder, "args" may be passed. These args depend an encoder requires during its preparation call.

    :param col_dtype: A data-type of a column specified
    :param col_name: The name of the column
    :param is_target: Whether the column is the target for prediction. If true, only certain possible feature representations are allowed, particularly for complex data types.
    :param problem_definition: The ``ProblemDefinition`` criteria; this populates specifics on how models and encoders may be trained.
    :param is_target_predicting_encoder:
    """  # noqa

    tss = problem_defintion.timeseries_settings
    encoder_lookup = {
        dtype.integer: "NumericEncoder",
        dtype.float: "NumericEncoder",
        dtype.binary: "BinaryEncoder",
        dtype.categorical: "CategoricalAutoEncoder"
        if statistical_analysis is None
        or len(statistical_analysis.histograms[col_name]) > 100
        else "OneHotEncoder",
        dtype.tags: "MultiHotEncoder",
        dtype.date: "DatetimeEncoder",
        dtype.datetime: "DatetimeEncoder",
        dtype.image: "Img2VecEncoder",
        dtype.rich_text: "PretrainedLangEncoder",
        dtype.short_text: "CategoricalAutoEncoder",
        dtype.quantity: "NumericEncoder",
        dtype.audio: "MFCCEncoder",
        dtype.num_array: "NumArrayEncoder",
        dtype.cat_array: "CatArrayEncoder",
        dtype.num_tsarray: "TimeSeriesEncoder",
        dtype.cat_tsarray: "TimeSeriesEncoder",
    }

    # If column is a target, only specific feature representations are allowed that enable supervised tasks
    target_encoder_lookup_override = {
        dtype.rich_text: "VocabularyEncoder",
        dtype.categorical: "OneHotEncoder",
    }

    # Assign a default encoder to each column.
    encoder_dict = {"module": encoder_lookup[col_dtype], "args": {}}

    # If the column is a target, ensure that the feature representation can enable supervised tasks
    if is_target:
        encoder_dict["args"] = {"is_target": "True"}

        if col_dtype in target_encoder_lookup_override:
            encoder_dict["module"] = target_encoder_lookup_override[col_dtype]

        if col_dtype in (dtype.categorical, dtype.binary):
            if problem_defintion.unbias_target:
                encoder_dict["args"][
                    "target_weights"
                ] = "$statistical_analysis.target_weights"
            if problem_defintion.target_weights is not None:
                encoder_dict["args"][
                    "target_weights"
                ] = problem_defintion.target_weights

        if col_dtype in (dtype.integer, dtype.float, dtype.num_array, dtype.num_tsarray):
            encoder_dict["args"][
                "positive_domain"
            ] = "$statistical_analysis.positive_domain"

    # Time-series representations require more advanced flags
    if tss.is_timeseries:
        gby = tss.group_by if tss.group_by is not None else []
        if col_name in tss.order_by:
            encoder_dict["module"] = "ArrayEncoder"
            encoder_dict["args"]["original_type"] = f'"{tss.target_type}"'
            encoder_dict["args"]["window"] = f"{tss.window}"

        if is_target:
            if col_dtype in [dtype.integer]:
                encoder_dict["args"]["grouped_by"] = f"{gby}"
                encoder_dict["module"] = "TsNumericEncoder"
            if col_dtype in [dtype.float]:
                encoder_dict["args"]["grouped_by"] = f"{gby}"
                encoder_dict["module"] = "TsNumericEncoder"
            if tss.horizon > 1:
                encoder_dict["args"]["grouped_by"] = f"{gby}"
                encoder_dict["args"]["timesteps"] = f"{tss.horizon}"
                if col_dtype in [dtype.num_tsarray]:
                    encoder_dict["module"] = "TsArrayNumericEncoder"
                elif col_dtype in [dtype.cat_tsarray]:
                    encoder_dict["module"] = "TsCatArrayEncoder"

        if "__mdb_ts_previous" in col_name or col_name in tss.historical_columns:
            encoder_dict["module"] = "TimeSeriesEncoder"
            encoder_dict["args"]["original_type"] = f'"{tss.target_type}"'
            encoder_dict["args"]["window"] = f"{tss.window}"

    # Set arguments for the encoder
    if encoder_dict["module"] == "PretrainedLangEncoder" and not is_target:
        encoder_dict["args"]["output_type"] = "$dtype_dict[$target]"

    if eval(encoder_dict["module"]).is_trainable_encoder:
        encoder_dict["args"]["stop_after"] = "$problem_definition.seconds_per_encoder"

    if is_target_predicting_encoder:
        encoder_dict["args"]["embed_mode"] = "False"
    return encoder_dict


def generate_json_ai(
    type_information: TypeInformation,
    statistical_analysis: StatisticalAnalysis,
    problem_definition: ProblemDefinition,
) -> JsonAI:
    """
    Given ``TypeInformation``, ``StatisticalAnalysis``, and the ``ProblemDefinition``, generate a JSON config file with the necessary elements of the ML pipeline populated.

    :param TypeInformation: Specifies what data types each column within the dataset are
    :param statistical_analysis:
    :param problem_definition: Specifies details of the model training/building procedure, as defined by ``ProblemDefinition``

    :returns: JSON-AI object with fully populated details of the ML pipeline
    """  # noqaexec
    exec(IMPORTS, globals())
    exec(IMPORT_EXTERNAL_DIRS, globals())
    target = problem_definition.target
    input_cols = []
    tss = problem_definition.timeseries_settings
    dtype_dict = type_information.dtypes
    for k in type_information.identifiers:
        if not (tss.is_timeseries and tss.group_by and k in tss.group_by):
            del dtype_dict[k]
    dependency_dict = {}

    for col_name, col_dtype in dtype_dict.items():
        if (
                (col_name not in type_information.identifiers
                 and col_dtype not in (dtype.invalid, dtype.empty)
                 and col_name != target)
                or
                (tss.group_by is not None and col_name in tss.group_by)
        ):
            if col_name != problem_definition.target:
                input_cols.append(col_name)

    is_target_predicting_encoder = False
    is_ts = problem_definition.timeseries_settings.is_timeseries
    imputers = []

    # Single text column classification
    if (
        len(input_cols) == 1
        and type_information.dtypes[input_cols[0]] in (dtype.rich_text)
        and type_information.dtypes[target] in (dtype.categorical, dtype.binary)
    ):
        is_target_predicting_encoder = True

    if is_target_predicting_encoder:
        submodels = [
            {
                "module": "Unit",
                "args": {
                    "target_encoder": "$encoders[self.target]",
                    "stop_after": "$problem_definition.seconds_per_mixer",
                },
            }
        ]
    else:
        submodels = [
            {
                "module": "Neural",
                "args": {
                    "fit_on_dev": True,
                    "stop_after": "$problem_definition.seconds_per_mixer",
                    "search_hyperparameters": True,
                },
            }
        ]

        if (not tss.is_timeseries or tss.horizon == 1) and dtype_dict[target] not in (dtype.num_array, dtype.cat_array):
            submodels.extend(
                [
                    {
                        "module": "LightGBM",
                        "args": {
                            "stop_after": "$problem_definition.seconds_per_mixer",
                            "fit_on_dev": True,
                        },
                    },
                    {
                        "module": "Regression",
                        "args": {
                            "stop_after": "$problem_definition.seconds_per_mixer",
                        },
                    },
                ]
            )
        elif tss.is_timeseries and tss.horizon > 1:
            submodels.extend(
                [
                    {
                        "module": "LightGBMArray",
                        "args": {
                            "fit_on_dev": True,
                            "stop_after": "$problem_definition.seconds_per_mixer",
                            "horizon": "$problem_definition.timeseries_settings.horizon",
                        },
                    }
                ]
            )

            if tss.use_previous_target and dtype_dict[target] in (dtype.integer, dtype.float, dtype.quantity):
                submodels.extend(
                    [
                        {
                            "module": "SkTime",
                            "args": {
                                "stop_after": "$problem_definition.seconds_per_mixer",
                                "horizon": "$problem_definition.timeseries_settings.horizon",
                            },
                        }
                    ]
                )

    model = {
        "module": "BestOf",
        "args": {
            "submodels": submodels,
            "args": "$pred_args",
            "accuracy_functions": "$accuracy_functions",
            "ts_analysis": "self.ts_analysis" if is_ts else None,
        }
    }

    num_ts_dtypes = (dtype.integer, dtype.float, dtype.quantity)
    if tss.is_timeseries and tss.horizon > 1:
        if dtype_dict[target] in num_ts_dtypes:
            dtype_dict[target] = dtype.num_tsarray
            problem_definition.anomaly_detection = True
        else:
            dtype_dict[target] = dtype.cat_tsarray
    elif tss.is_timeseries and dtype_dict[target] in num_ts_dtypes:
        problem_definition.anomaly_detection = True

    encoders = {
        target: lookup_encoder(
            dtype_dict[target],
            target,
            True,
            problem_definition,
            False,
            statistical_analysis,
        )
    }

    for col in input_cols:
        encoders[col] = lookup_encoder(
            dtype_dict[col],
            col,
            False,
            problem_definition,
            is_target_predicting_encoder,
            statistical_analysis,
        )

    # Decide on the accuracy functions to use
    output_dtype = dtype_dict[target]
    if output_dtype in [
        dtype.integer,
        dtype.float,
        dtype.date,
        dtype.datetime,
        dtype.quantity,
    ]:
        accuracy_functions = ["r2_score"]
    elif output_dtype in [dtype.categorical, dtype.tags, dtype.binary]:
        accuracy_functions = ["balanced_accuracy_score"]
    elif output_dtype in (dtype.num_array, dtype.num_tsarray):
        accuracy_functions = ["evaluate_num_array_accuracy"]
    elif output_dtype in (dtype.cat_array, dtype.cat_tsarray):
        accuracy_functions = ["evaluate_cat_array_accuracy"]
    else:
        raise Exception(
            f"Please specify a custom accuracy function for output type {output_dtype}"
        )

    if is_ts:
        if output_dtype in [dtype.integer, dtype.float]:
            accuracy_functions = ["evaluate_num_array_accuracy"]  # forces this acc fn for t+1 time series forecasters

        if output_dtype in (dtype.integer, dtype.float, dtype.num_tsarray):
            imputers.append({"module": "NumericalImputer",
                             "args": {
                                 "value": "'zero'",
                                 "target": f"'{target}'"}}
                            )
        elif output_dtype in [dtype.categorical, dtype.tags, dtype.binary, dtype.cat_tsarray]:
            imputers.append({"module": "CategoricalImputer",
                             "args": {
                                 "value": "'mode'",
                                 "target": f"'{target}'"}}
                            )

    if problem_definition.time_aim is None:
        # 5 days
        problem_definition.time_aim = 3 * 24 * 3600

    # Encoders are assigned 1/3 of the time unless a user overrides this (equal time per encoder)
    if problem_definition.seconds_per_encoder is None:
        nr_trainable_encoders = len(
            [
                x
                for x in encoders.values()
                if eval(x["module"]).is_trainable_encoder
            ]
        )
        if nr_trainable_encoders > 0:
            problem_definition.seconds_per_encoder = 0.33 * problem_definition.time_aim / nr_trainable_encoders

    # Mixers are assigned 1/3 of the time aim (or 2/3 if there are no trainable encoders )\
    # unless a user overrides this (equal time per mixer)
    if problem_definition.seconds_per_mixer is None:
        if problem_definition.seconds_per_encoder is None:
            problem_definition.seconds_per_mixer = 0.66 * problem_definition.time_aim / len(model['args']['submodels'])
        else:
            problem_definition.seconds_per_mixer = 0.33 * problem_definition.time_aim / len(model['args']['submodels'])

    return JsonAI(
        cleaner=None,
        splitter=None,
        analyzer=None,
        explainer=None,
        encoders=encoders,
        imputers=imputers,
        dtype_dict=dtype_dict,
        dependency_dict=dependency_dict,
        model=model,
        problem_definition=problem_definition,
        identifiers=type_information.identifiers,
        timeseries_transformer=None,
        timeseries_analyzer=None,
        accuracy_functions=accuracy_functions,
    )


def _merge_implicit_values(field: dict, implicit_value: dict) -> dict:
    """
    Helper function for `_populate_implicit_field`.
    Takes a user-defined field along with its implicit value, and merges them together.

    :param field: JsonAI field with user-defined parameters.
    :param implicit_value: implicit values for the field.
    :return: original field with implicit values merged into it.
    """
    exec(IMPORTS, globals())
    exec(IMPORT_EXTERNAL_DIRS, globals())
    module = eval(field["module"])

    if inspect.isclass(module):
        args = list(inspect.signature(module.__init__).parameters.keys())[1:]
    else:
        args = module.__code__.co_varnames

    for arg in args:
        if "args" not in field:
            field["args"] = implicit_value["args"]
        else:
            if arg not in field["args"]:
                if arg in implicit_value["args"]:
                    field["args"][arg] = implicit_value["args"][arg]

    return field


def _populate_implicit_field(
    json_ai: JsonAI, field_name: str, implicit_value: dict, is_timeseries: bool
) -> None:
    """
    Populate the implicit field of the JsonAI, either by filling it in entirely if missing, or by introspecting the class or function and assigning default values to the args in it's signature that are in the implicit default but haven't been populated by the user

    :params: json_ai: ``JsonAI`` object that describes the ML pipeline that may not have every detail fully specified.
    :params: field_name: Name of the field the implicit field in ``JsonAI``
    :params: implicit_value: The dictionary containing implicit values for the module and arg in the field
    :params: is_timeseries: Whether or not this is a timeseries problem

    :returns: nothing, this method mutates the respective field of the ``JsonAI`` object it receives
    """  # noqa
    # These imports might be slow, in which case the only <easy> solution is to line this code
    field = json_ai.__getattribute__(field_name)
    if field is None:
        # This if is to only populated timeseries-specific implicit fields for implicit problems
        if is_timeseries or field_name not in (
            "timeseries_analyzer",
            "timeseries_transformer",
        ):
            field = implicit_value

    # If the user specified one or more subfields in a field that's a list
    # Populate them with implicit arguments form the implicit values from that subfield
    elif isinstance(field, list) and isinstance(implicit_value, list):
        for i in range(len(field)):
            sub_field_implicit = [
                x for x in implicit_value if x["module"] == field[i]["module"]
            ]
            if len(sub_field_implicit) == 1:
                field[i] = _merge_implicit_values(field[i], sub_field_implicit[0])
        for sub_field_implicit in implicit_value:
            if (
                len([x for x in field if x["module"] == sub_field_implicit["module"]])
                == 0
            ):
                field.append(sub_field_implicit)
    # If the user specified the field, add implicit arguments which we didn't specify
    else:
        field = _merge_implicit_values(field, implicit_value)
    json_ai.__setattr__(field_name, field)


def _add_implicit_values(json_ai: JsonAI) -> JsonAI:
    """
    To enable brevity in writing, auto-generate the "unspecified/missing" details required in the ML pipeline.

    :params: json_ai: ``JsonAI`` object that describes the ML pipeline that may not have every detail fully specified.

    :returns: ``JSONAI`` object with all necessary parameters that were previously left unmentioned filled in.
    """
    problem_definition = json_ai.problem_definition
    tss = problem_definition.timeseries_settings
    is_ts = tss.is_timeseries

    # Add implicit arguments
    # @TODO: Consider removing once we have a proper editor in studio
    mixers = json_ai.model['args']['submodels']
    for i in range(len(mixers)):
        if mixers[i]["module"] == "Unit":
            pass

        elif mixers[i]["module"] == "Neural":
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )
            mixers[i]["args"]["target"] = mixers[i]["args"].get("target", "$target")
            mixers[i]["args"]["dtype_dict"] = mixers[i]["args"].get(
                "dtype_dict", "$dtype_dict"
            )
            mixers[i]["args"]["timeseries_settings"] = mixers[i]["args"].get(
                "timeseries_settings", "$problem_definition.timeseries_settings"
            )
            mixers[i]["args"]["net"] = mixers[i]["args"].get(
                "net",
                '"DefaultNet"'
                if not tss.is_timeseries or not tss.use_previous_target
                else '"ArNet"',
            )

        elif mixers[i]["module"] == "LightGBM":
            mixers[i]["args"]["target"] = mixers[i]["args"].get("target", "$target")
            mixers[i]["args"]["dtype_dict"] = mixers[i]["args"].get(
                "dtype_dict", "$dtype_dict"
            )
            mixers[i]["args"]["input_cols"] = mixers[i]["args"].get(
                "input_cols", "$input_cols"
            )
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )
            mixers[i]["args"]["use_optuna"] = True

        elif mixers[i]["module"] == "Regression":
            mixers[i]["args"]["target"] = mixers[i]["args"].get("target", "$target")
            mixers[i]["args"]["dtype_dict"] = mixers[i]["args"].get(
                "dtype_dict", "$dtype_dict"
            )
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )

        elif mixers[i]["module"] == "LightGBMArray":
            mixers[i]["args"]["target"] = mixers[i]["args"].get("target", "$target")
            mixers[i]["args"]["dtype_dict"] = mixers[i]["args"].get(
                "dtype_dict", "$dtype_dict"
            )
            mixers[i]["args"]["input_cols"] = mixers[i]["args"].get(
                "input_cols", "$input_cols"
            )
            mixers[i]["args"]["target_encoder"] = mixers[i]["args"].get(
                "target_encoder", "$encoders[self.target]"
            )
            if "horizon" not in mixers[i]["args"]:
                mixers[i]["args"]["horizon"] = "$problem_definition.timeseries_settings.horizon"

        elif mixers[i]["module"] in ("SkTime", "ProphetMixer"):
            mixers[i]["args"]["target"] = mixers[i]["args"].get("target", "$target")
            mixers[i]["args"]["dtype_dict"] = mixers[i]["args"].get(
                "dtype_dict", "$dtype_dict"
            )
            mixers[i]["args"]["ts_analysis"] = mixers[i]["args"].get(
                "ts_analysis", "$ts_analysis"
            )
            if "horizon" not in mixers[i]["args"]:
                mixers[i]["args"]["horizon"] = "$problem_definition.timeseries_settings.horizon"

            # enforce fit_on_all if this mixer is specified
            problem_definition.fit_on_all = True

        if "stop_after" not in mixers[i]["args"]:
            mixers[i]["args"]["stop_after"] = "$problem_definition.seconds_per_mixer"

    json_ai.model["args"]["target"] = json_ai.model["args"].get("target", "$target")
    json_ai.model["args"]["data"] = json_ai.model["args"].get("data", "encoded_test_data")
    json_ai.model["args"]["mixers"] = json_ai.model["args"].get("mixers", "$mixers")

    for name in json_ai.encoders:
        if name not in json_ai.dependency_dict:
            json_ai.dependency_dict[name] = []

    # Add "hidden" fields
    hidden_fields = {
        "cleaner": {
            "module": "cleaner",
            "args": {
                "pct_invalid": "$problem_definition.pct_invalid",
                "identifiers": "$identifiers",
                "data": "data",
                "dtype_dict": "$dtype_dict",
                "target": "$target",
                "mode": "$mode",
                "imputers": "$imputers",
                "timeseries_settings": "$problem_definition.timeseries_settings",
                "anomaly_detection": "$problem_definition.anomaly_detection",
            },
        },
        "splitter": {
            "module": "splitter",
            "args": {
                "tss": "$problem_definition.timeseries_settings",
                "data": "data",
                "seed": "$problem_definition.seed_nr",
                "target": "$target",
                "dtype_dict": "$dtype_dict",
                "pct_train": 0.8,
                "pct_dev": 0.1,
                "pct_test": 0.1,
            },
        },
        "analyzer": {
            "module": "model_analyzer",
            "args": {
                "stats_info": "$statistical_analysis",
                "tss": "$problem_definition.timeseries_settings",
                "accuracy_functions": "$accuracy_functions",
                "predictor": "$ensemble",
                "data": "encoded_test_data",
                "train_data": "encoded_train_data",
                "target": "$target",
                "dtype_dict": "$dtype_dict",
                "analysis_blocks": "$analysis_blocks",
                "ts_analysis": "$ts_analysis" if is_ts else None,
            },
        },
        "explainer": {
            "module": "explain",
            "args": {
                "timeseries_settings": "$problem_definition.timeseries_settings",
                "positive_domain": "$statistical_analysis.positive_domain",
                "anomaly_detection": "$problem_definition.anomaly_detection",
                "data": "data",
                "encoded_data": "encoded_data",
                "predictions": "df",
                "analysis": "$runtime_analyzer",
                "ts_analysis": "$ts_analysis" if is_ts else None,
                "target_name": "$target",
                "target_dtype": "$dtype_dict[self.target]",
                "explainer_blocks": "$analysis_blocks",
                "pred_args": "$pred_args",
            },
        },
        "analysis_blocks": [
            {
                "module": "ICP",
                "args": {
                    "fixed_significance": None,
                    "confidence_normalizer": False,
                    "positive_domain": "$statistical_analysis.positive_domain",
                },
            },
            {
                "module": "AccStats",
                "args": {"deps": ["ICP"]},
            },
            {
                "module": "ConfStats",
                "args": {"deps": ["ICP"]},
            },
        ] if problem_definition.use_default_analysis else [],
        "timeseries_transformer": {
            "module": "transform_timeseries",
            "args": {
                "timeseries_settings": "$problem_definition.timeseries_settings",
                "data": "data",
                "dtype_dict": "$dtype_dict",
                "target": "$target",
                "mode": "$mode",
            },
        },
        "timeseries_analyzer": {
            "module": "timeseries_analyzer",
            "args": {
                "timeseries_settings": "$problem_definition.timeseries_settings",
                "data": "data",
                "dtype_dict": "$dtype_dict",
                "target": "$target",
            },
        },
    }

    for field_name, implicit_value in hidden_fields.items():
        _populate_implicit_field(json_ai, field_name, implicit_value, tss.is_timeseries)

    return json_ai


def code_from_json_ai(json_ai: JsonAI) -> str:
    """
    Generates a custom ``PredictorInterface`` given the specifications from ``JsonAI`` object.

    :param json_ai: ``JsonAI`` object with fully specified parameters

    :returns: Automated syntax of the ``PredictorInterface`` object.
    """
    json_ai = deepcopy(json_ai)
    # ----------------- #
    # Fill in any missing values
    json_ai = _add_implicit_values(json_ai)

    # ----------------- #

    # Instantiate data types
    dtype_dict = {}

    for k in json_ai.dtype_dict:
        if json_ai.dtype_dict[k] not in (dtype.invalid, dtype.empty):
            dtype_dict[k] = json_ai.dtype_dict[k]

    # Populate imputers
    imputer_dict = {}
    if json_ai.imputers:
        for imputer in json_ai.imputers:
            imputer_dict[imputer['args']['target'].replace('\'', '').replace('\"', '')] = call(imputer)
    json_ai.imputers = imputer_dict
    imputers = inline_dict(json_ai.imputers)

    # Populate encoders
    encoder_dict = {}
    for col_name, encoder in json_ai.encoders.items():
        encoder_dict[col_name] = call(encoder)

    # Populate time-series specific details
    tss = json_ai.problem_definition.timeseries_settings
    if tss.is_timeseries and tss.use_previous_target:
        col_name = f"__mdb_ts_previous_{json_ai.problem_definition.target}"
        target_type = json_ai.dtype_dict[json_ai.problem_definition.target]
        json_ai.problem_definition.timeseries_settings.target_type = target_type
        encoder_dict[col_name] = call(
            lookup_encoder(
                target_type,
                col_name,
                False,
                json_ai.problem_definition,
                False,
                None,
            )
        )

        dtype_dict[col_name] = target_type
        # @TODO: Is populating the json_ai at this stage even necessary?
        json_ai.encoders[col_name] = encoder_dict[col_name]
        json_ai.dtype_dict[col_name] = target_type
        json_ai.dependency_dict[col_name] = []

    # ----------------- #

    input_cols = [x.replace("'", "\\'").replace('"', '\\"') for x in json_ai.encoders
                  if x != json_ai.problem_definition.target]
    input_cols = ",".join([f"""'{name}'""" for name in input_cols])

    # ----------------- #
    # Time-series specific code blocks
    # ----------------- #

    ts_transform_code = ""
    ts_analyze_code = None
    ts_encoder_code = ""
    if json_ai.timeseries_transformer is not None:
        ts_transform_code = f"""
log.info('Transforming timeseries data')
data = {call(json_ai.timeseries_transformer)}
"""
        ts_analyze_code = f"""
self.ts_analysis = {call(json_ai.timeseries_analyzer)}
"""
    # @TODO: set these kwargs/properties in the json ai construction (if possible)
    if json_ai.timeseries_analyzer is not None:
        ts_encoder_code = """
if encoder.is_timeseries_encoder:
    kwargs['ts_analysis'] = self.ts_analysis
"""

    if json_ai.problem_definition.timeseries_settings.is_timeseries:
        ts_target_code = """
if encoder.is_target:
    encoder.normalizers = self.ts_analysis['target_normalizers']
    encoder.group_combinations = self.ts_analysis['group_combinations']
"""
    else:
        ts_target_code = ""

    # ----------------- #
    # Statistical Analysis Body
    # ----------------- #

    analyze_data_body = f"""
self.statistical_analysis = lightwood.data.statistical_analysis(data,
                                                                self.dtype_dict,
                                                                {json_ai.identifiers},
                                                                self.problem_definition)

# Instantiate post-training evaluation
self.analysis_blocks = [{', '.join([call(block) for block in json_ai.analysis_blocks])}]
    """

    analyze_data_body = align(analyze_data_body, 2)

    # ----------------- #
    # Pre-processing Body
    # ----------------- #

    clean_body = f"""
log.info('Cleaning the data')
self.imputers = {imputers}
data = {call(json_ai.cleaner)}

# Time-series blocks
{ts_transform_code}
"""
    if ts_analyze_code is not None:
        clean_body += f"""
if self.mode != 'predict':
{align(ts_analyze_code,1)}
"""

    clean_body += '\nreturn data'

    clean_body = align(clean_body, 2)

    # ----------------- #
    # Train-Test Splitter Body
    # ----------------- #

    split_body = f"""
log.info("Splitting the data into train/test")
train_test_data = {call(json_ai.splitter)}

return train_test_data
    """

    split_body = align(split_body, 2)

    # ----------------- #
    # Prepare features Body
    # ----------------- #

    prepare_body = f"""
self.mode = 'train'

if self.statistical_analysis is None:
    raise Exception("Please run analyze_data first")

# Column to encoder mapping
self.encoders = {inline_dict(encoder_dict)}

# Prepare the training + dev data
concatenated_train_dev = pd.concat([data['train'], data['dev']])

encoder_prepping_dict = {{}}

# Prepare encoders that do not require learned strategies
for col_name, encoder in self.encoders.items():
    if col_name != self.target and not encoder.is_trainable_encoder:
        encoder_prepping_dict[col_name] = [encoder, concatenated_train_dev[col_name], 'prepare']

# Setup parallelization
parallel_prepped_encoders = mut_method_call(encoder_prepping_dict)
for col_name, encoder in parallel_prepped_encoders.items():
    self.encoders[col_name] = encoder

# Prepare the target
if self.target not in parallel_prepped_encoders:
    if self.encoders[self.target].is_trainable_encoder:
        self.encoders[self.target].prepare(data['train'][self.target], data['dev'][self.target])
    else:
        self.encoders[self.target].prepare(pd.concat([data['train'], data['dev']])[self.target])

# Prepare any non-target encoders that are learned
for col_name, encoder in self.encoders.items():
    if col_name != self.target and encoder.is_trainable_encoder:
        priming_data = pd.concat([data['train'], data['dev']])
        kwargs = {{}}
        if self.dependencies[col_name]:
            kwargs['dependency_data'] = {{}}
            for col in self.dependencies[col_name]:
                kwargs['dependency_data'][col] = {{
                    'original_type': self.dtype_dict[col],
                    'data': priming_data[col]
                }}
            {align(ts_encoder_code, 3)}

        # If an encoder representation requires the target, provide priming data
        if hasattr(encoder, 'uses_target'):
            kwargs['encoded_target_values'] = self.encoders[self.target].encode(priming_data[self.target])

        encoder.prepare(data['train'][col_name], data['dev'][col_name], **kwargs)

    {align(ts_target_code, 1)}
"""
    prepare_body = align(prepare_body, 2)

    # ----------------- #
    # Featurize Data Body
    # ----------------- #

    feature_body = f"""
log.info('Featurizing the data')

feature_data = {{ key: EncodedDs(self.encoders, data, self.target) for key, data in split_data.items() if key != "stratified_on"}}

return feature_data

"""  # noqa

    feature_body = align(feature_body, 2)

    # ----------------- #
    # Fit Mixer Body
    # ----------------- #

    fit_body = f"""
self.mode = 'train'

# --------------- #
# Extract data
# --------------- #
# Extract the featurized data into train/dev/test
encoded_train_data = enc_data['train']
encoded_dev_data = enc_data['dev']
encoded_test_data = enc_data['test']

log.info('Training the mixers')

# --------------- #
# Fit Models
# --------------- #
# Assign list of mixers
self.mixers = [{', '.join([call(x) for x in json_ai.model["args"]["submodels"]])}]

# Train mixers
trained_mixers = []
for mixer in self.mixers:
    try:
        self.fit_mixer(mixer, encoded_train_data, encoded_dev_data)
        trained_mixers.append(mixer)
    except Exception as e:
        log.warning(f'Exception: {{e}} when training mixer: {{mixer}}')
        if {json_ai.problem_definition.strict_mode} and mixer.stable:
            raise e

# Update mixers to trained versions
self.mixers = trained_mixers

# --------------- #
# Create Ensembles
# --------------- #
log.info('Ensembling the mixer')
# Create an ensemble of mixers to identify best performing model
self.pred_args = PredictionArguments()
# Dirty hack
self.ensemble = {call(json_ai.model)}
self.supports_proba = self.ensemble.supports_proba
"""
    fit_body = align(fit_body, 2)

    # ----------------- #
    # Analyze Ensemble Body
    # ----------------- #

    analyze_ensemble = f"""

# --------------- #
# Extract data
# --------------- #
# Extract the featurized data into train/dev/test
encoded_train_data = enc_data['train']
encoded_dev_data = enc_data['dev']
encoded_test_data = enc_data['test']

# --------------- #
# Analyze Ensembles
# --------------- #
log.info('Analyzing the ensemble of mixers')
self.model_analysis, self.runtime_analyzer = {call(json_ai.analyzer)}
"""
    analyze_ensemble = align(analyze_ensemble, 2)

    # ----------------- #
    # Adjust Ensemble Body
    # ----------------- #

    adjust_body = f"""
self.mode = 'train'

# --------------- #
# Prepare data
# --------------- #
if dev_data is None:
    data = train_data if isinstance(train_data, pd.DataFrame) else train_data.data_frame
    split = splitter(data,
        self.problem_definition.timeseries_settings,
        self.dtype_dict,
        self.problem_definition.seed_nr,
        pct_train=0.8,
        pct_dev=0.2,
        pct_test=0,
        target=self.target)
    train_data = split['train']
    dev_data = split['dev']

if isinstance(dev_data, pd.DataFrame):
    dev_data = EncodedDs(self.encoders, dev_data, self.target)

if isinstance(train_data, pd.DataFrame):
    train_data = EncodedDs(self.encoders, train_data, self.target)

# --------------- #
# Update/Adjust Mixers
# --------------- #
log.info('Updating the mixers')

for mixer in self.mixers:
    mixer.partial_fit(train_data, dev_data)
"""  # noqa

    adjust_body = align(adjust_body, 2)

    # ----------------- #
    # Learn Body
    # ----------------- #

    learn_body = """
self.mode = 'train'
n_phases = 8 if self.problem_definition.fit_on_all else 7

# Perform stats analysis
log.info(f'[Learn phase 1/{n_phases}] - Statistical analysis')
self.analyze_data(data)

# Pre-process the data
log.info(f'[Learn phase 2/{n_phases}] - Data preprocessing')
data = self.preprocess(data)

# Create train/test (dev) split
log.info(f'[Learn phase 3/{n_phases}] - Data splitting')
train_dev_test = self.split(data)

# Prepare encoders
log.info(f'[Learn phase 4/{n_phases}] - Preparing encoders')
self.prepare(train_dev_test)

# Create feature vectors from data
log.info(f'[Learn phase 5/{n_phases}] - Feature generation')
enc_train_test = self.featurize(train_dev_test)

# Prepare mixers
log.info(f'[Learn phase 6/{n_phases}] - Mixer training')
self.fit(enc_train_test)

# Analyze the ensemble
log.info(f'[Learn phase 7/{n_phases}] - Ensemble analysis')
self.analyze_ensemble(enc_train_test)

# ------------------------ #
# Enable model partial fit AFTER it is trained and evaluated for performance with the appropriate train/dev/test splits.
# This assumes the predictor could continuously evolve, hence including reserved testing data may improve predictions.
# SET `json_ai.problem_definition.fit_on_all=False` TO TURN THIS BLOCK OFF.

# Update the mixers with partial fit
if self.problem_definition.fit_on_all:

    log.info(f'[Learn phase 8/{n_phases}] - Adjustment on validation requested')
    self.adjust(enc_train_test["test"], ConcatedEncodedDs([enc_train_test["train"], enc_train_test["dev"]]))

"""
    learn_body = align(learn_body, 2)
    # ----------------- #
    # Predict Body
    # ----------------- #

    predict_body = f"""
self.mode = 'predict'
n_phases = 3 if self.pred_args.all_mixers else 4

if len(data) == 0:
    raise Exception("Empty input, aborting prediction. Please try again with some input data.")

log.info(f'[Predict phase 1/{{n_phases}}] - Data preprocessing')
if self.problem_definition.ignore_features:
    log.info(f'Dropping features: {{self.problem_definition.ignore_features}}')
    data = data.drop(columns=self.problem_definition.ignore_features, errors='ignore')
for col in self.input_cols:
    if col not in data.columns:
        data[col] = [None] * len(data)

# Pre-process the data
data = self.preprocess(data)

# Featurize the data
log.info(f'[Predict phase 2/{{n_phases}}] - Feature generation')
encoded_ds = self.featurize({{"predict_data": data}})["predict_data"]
encoded_data = encoded_ds.get_encoded_data(include_target=False)

log.info(f'[Predict phase 3/{{n_phases}}] - Calling ensemble')
self.pred_args = PredictionArguments.from_dict(args)
df = self.ensemble(encoded_ds, args=self.pred_args)

if self.pred_args.all_mixers:
    return df
else:
    log.info(f'[Predict phase 4/{{n_phases}}] - Analyzing output')
    insights, global_insights = {call(json_ai.explainer)}
    return insights
"""

    predict_body = align(predict_body, 2)

    predictor_code = f"""
{IMPORTS}
{IMPORT_EXTERNAL_DIRS}

class Predictor(PredictorInterface):
    target: str
    mixers: List[BaseMixer]
    encoders: Dict[str, BaseEncoder]
    ensemble: BaseEnsemble
    mode: str

    def __init__(self):
        seed({json_ai.problem_definition.seed_nr})
        self.target = '{json_ai.problem_definition.target}'
        self.mode = 'inactive'
        self.problem_definition = ProblemDefinition.from_dict({json_ai.problem_definition.to_dict()})
        self.accuracy_functions = {json_ai.accuracy_functions}
        self.identifiers = {json_ai.identifiers}
        self.dtype_dict = {inline_dict(dtype_dict)}

        # Any feature-column dependencies
        self.dependencies = {inline_dict(json_ai.dependency_dict)}

        self.input_cols = [{input_cols}]

        # Initial stats analysis
        self.statistical_analysis = None
        self.runtime_log = dict()

    @timed
    def analyze_data(self, data: pd.DataFrame) -> None:
        # Perform a statistical analysis on the unprocessed data
{analyze_data_body}

    @timed
    def preprocess(self, data: pd.DataFrame) -> pd.DataFrame:
        # Preprocess and clean data
{clean_body}

    @timed
    def split(self, data: pd.DataFrame) -> Dict[str, pd.DataFrame]:
        # Split the data into training/testing splits
{split_body}

    @timed
    def prepare(self, data: Dict[str, pd.DataFrame]) -> None:
        # Prepare encoders to featurize data
{prepare_body}

    @timed
    def featurize(self, split_data: Dict[str, pd.DataFrame]):
        # Featurize data into numerical representations for models
{feature_body}

    @timed
    def fit(self, enc_data: Dict[str, pd.DataFrame]) -> None:
        # Fit predictors to estimate target
{fit_body}

    @timed
    def fit_mixer(self, mixer, encoded_train_data, encoded_dev_data) -> None:
        mixer.fit(encoded_train_data, encoded_dev_data)

    @timed
    def analyze_ensemble(self, enc_data: Dict[str, pd.DataFrame]) -> None:
        # Evaluate quality of fit for the ensemble of mixers
{analyze_ensemble}

    @timed
    def learn(self, data: pd.DataFrame) -> None:
        if self.problem_definition.ignore_features:
            log.info(f'Dropping features: {{self.problem_definition.ignore_features}}')
            data = data.drop(columns=self.problem_definition.ignore_features, errors='ignore')
{learn_body}

    @timed
    def adjust(self, train_data: Union[EncodedDs, ConcatedEncodedDs, pd.DataFrame],
        dev_data: Optional[Union[EncodedDs, ConcatedEncodedDs, pd.DataFrame]] = None) -> None:
        # Update mixers with new information
{adjust_body}

    @timed
    def predict(self, data: pd.DataFrame, args: Dict = {{}}) -> pd.DataFrame:
{predict_body}
"""

    try:
        import black
    except Exception:
        black = None

    if black is not None:
        log.info('Unable to import black formatter, predictor code might be a bit ugly.')
        predictor_code = black.format_str(predictor_code, mode=black.FileMode())

    return predictor_code


def validate_json_ai(json_ai: JsonAI) -> bool:
    """
    Checks the validity of a ``JsonAI`` object

    :param json_ai: A ``JsonAI`` object

    :returns: Whether the JsonAI is valid, i.e. doesn't contain prohibited values, unknown values and can be turned into code.
    """ # noqa
    from lightwood.api.high_level import predictor_from_code, code_from_json_ai

    try:
        predictor_from_code(code_from_json_ai(json_ai))
        return True
    except Exception:
        return False



from itertools import product
from lightwood.api.dtype import dtype
from lightwood.encoder.helpers import MinMaxNormalizer, CatNormalizer
from lightwood.helpers.general import get_group_matches


def generate_target_group_normalizers(data):
    """
    Helper function called from data_source. It generates and fits all needed normalizers for a target variable
    based on its grouped entities.
    :param data:
    :return: modified data with dictionary with normalizers for said target variable based on some grouped-by columns
    """
    normalizers = {}
    group_combinations = []

    # categorical normalizers
    if data['original_type'] in (dtype.categorical, dtype.binary, dtype.cat_tsarray):
        normalizers['__default'] = CatNormalizer()
        normalizers['__default'].prepare(data['data'])
        group_combinations.append('__default')

    # numerical normalizers, here we spawn one per each group combination
    else:
        if data['original_type'] == dtype.num_tsarray:
            data['data'] = data['data'].reshape(-1, 1).astype(float)

        all_group_combinations = list(product(*[set(x) for x in data['group_info'].values()]))
        for combination in all_group_combinations:
            if combination != ():
                combination = tuple(combination)
                _, subset = get_group_matches(data, combination)
                if subset.size > 0:
                    normalizers[combination] = MinMaxNormalizer(combination=combination)
                    normalizers[combination].prepare(subset)
                    group_combinations.append(combination)

        # ...plus a default one, used at inference time and fitted with all training data
        normalizers['__default'] = MinMaxNormalizer()
        normalizers['__default'].prepare(data['data'])
        group_combinations.append('__default')

    data['target_normalizers'] = normalizers
    data['group_combinations'] = group_combinations

    return data



from typing import Iterable, List
from sklearn.metrics import r2_score as sk_r2_score
from sklearn.metrics import f1_score as sk_f1_score
from sklearn.metrics import recall_score as sk_recall_score
from sklearn.metrics import precision_score as sk_precision_score


def to_binary(y: Iterable) -> List[int]:
    try:
        y_binarized = []
        for ele in y:
            if str(ele).lower() == 'true':
                y_binarized.append(1)
            elif str(ele).lower() == 'false':
                y_binarized.append(0)
            else:
                y_binarized.append(int(ele))

        assert len(set(y_binarized)) < 3
        assert 1 in y_binarized or 0 in y_binarized
    except Exception:
        raise Exception('To use precision, recall or f1 please make sure your target consists only of 1s and 0s')
    return y_binarized


def f1_score(y_true, y_pred) -> float:
    return sk_f1_score(to_binary(y_true), to_binary(y_pred))


def recall_score(y_true, y_pred) -> float:
    return sk_recall_score(to_binary(y_true), to_binary(y_pred))


def precision_score(y_true, y_pred) -> float:
    return sk_precision_score(to_binary(y_true), to_binary(y_pred))


def r2_score(y_true, y_pred) -> float:
    """ Wrapper for sklearn R2 score, lower capped between 0 and 1"""
    acc = sk_r2_score(y_true, y_pred)
    # Cap at 0
    if acc < 0:
        acc = 0
    # Guard against overflow (> 1 means overflow of negative score)
    if acc > 1:
        acc = 0

    return acc



from typing import List, Optional

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from lightwood.helpers.numeric import is_nan_numeric
from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs
from lightwood.helpers.general import evaluate_accuracy
from lightwood import dtype


class WeightedMeanEnsemble(BaseEnsemble):
    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, args: PredictionArguments,
                 dtype_dict: dict, accuracy_functions, ts_analysis: Optional[dict] = None) -> None:
        super().__init__(target, mixers, data)
        if dtype_dict[target] not in (dtype.float, dtype.integer, dtype.quantity):
            raise Exception(
                f'This ensemble can only be used regression problems! Got target dtype {dtype_dict[target]} instead!')

        score_list = []
        for _, mixer in enumerate(mixers):
            score_dict = evaluate_accuracy(
                data.data_frame,
                mixer(data, args)['prediction'],
                target,
                accuracy_functions,
                ts_analysis=ts_analysis
            )
            avg_score = np.mean(list(score_dict.values()))
            log.info(f'Mixer: {type(mixer).__name__} got accuracy: {avg_score}')

            if is_nan_numeric(avg_score):
                log.warning(f'Could not compute a valid accuracy for mixer: {type(mixer).__name__}, \
                              functions: {accuracy_functions}, yielded invalid average score {avg_score}, \
                              resetting that to -pow(2,63) instead.')
                avg_score = -pow(2, 63)

            score_list.append(avg_score)

        self.weights = self.accuracies_to_weights(np.array(score_list))

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        df = pd.DataFrame()
        for mixer in self.mixers:
            df[f'__mdb_mixer_{type(mixer).__name__}'] = mixer(ds, args=args)['prediction']

        avg_predictions_df = df.apply(lambda x: np.average(x, weights=self.weights), axis='columns')
        return pd.DataFrame(avg_predictions_df, columns=['prediction'])

    def accuracies_to_weights(self, x: np.array) -> np.array:
        # Converts accuracies to weights using the softmax function.
        e_x = np.exp(x - np.max(x))
        return e_x / e_x.sum()



import unittest

from lightwood.encoder.text.vocab import VocabularyEncoder


class TestVocabularyEncoder(unittest.TestCase):
    def test_encode_decode(self):
        self.maxDiff = None
        sentences = [
            'Consequuntur magni dolores eos qui ratione voluptatem sequi nesciunt. Neque porro quisquam est, qui dolorem ipsum quia dolor sit amet, consectetur, adipisci velit, sed quia non numquam eiu.', # noqa
            'At vero eos et accusamus et iusto odio dignissimos ducimus qui blanditiis praesentium voluptatum deleniti atque corrupti quos dolores et quas molestias excepturi sint occaecati cupiditate non provident', # noqa
            'Itaque earum rerum hic tenetur a sapiente delectus, ut aut reiciendis voluptatibus maiores alias? consequatur aut perferendis doloribus asperiores repellat.'] # noqa

        sentences = [x.lower() for x in sentences]

        encoder = VocabularyEncoder()
        encoder.prepare(sentences)

        for sentence in sentences:
            encoded = encoder.encode([sentence])
            decoded = encoder.decode(encoded)
            self.assertEqual(decoded[0], sentence)



from typing import Dict, List, Tuple, Optional

from lightwood.helpers.log import log
from lightwood.api import dtype
from lightwood.ensemble import BaseEnsemble
from lightwood.analysis.base import BaseAnalysisBlock
from lightwood.data.encoded_ds import EncodedDs
from lightwood.encoder.text.pretrained import PretrainedLangEncoder
from lightwood.api.types import ModelAnalysis, StatisticalAnalysis, TimeseriesSettings, PredictionArguments


def model_analyzer(
    predictor: BaseEnsemble,
    data: EncodedDs,
    train_data: EncodedDs,
    stats_info: StatisticalAnalysis,
    target: str,
    tss: TimeseriesSettings,
    dtype_dict: Dict[str, str],
    accuracy_functions,
    ts_analysis: Dict,
    analysis_blocks: Optional[List[BaseAnalysisBlock]] = []
) -> Tuple[ModelAnalysis, Dict[str, object]]:
    """
    Analyses model on a validation subset to evaluate accuracy, estimate feature importance and generate a
    calibration model to estimating confidence in future predictions.

    Additionally, any user-specified analysis blocks (see class `BaseAnalysisBlock`) are also called here.

    :return:
    runtime_analyzer: This dictionary object gets populated in a sequential fashion with data generated from
    any `.analyze()` block call. This dictionary object is stored in the predictor itself, and used when
    calling the `.explain()` method of all analysis blocks when generating predictions.

    model_analysis: `ModelAnalysis` object that contains core analysis metrics, not necessarily needed when predicting.
    """

    runtime_analyzer = {}
    data_type = dtype_dict[target]

    # retrieve encoded data representations
    encoded_train_data = train_data
    encoded_val_data = data
    data = encoded_val_data.data_frame
    input_cols = list([col for col in data.columns if col != target])

    # predictive task
    is_numerical = data_type in (dtype.integer, dtype.float, dtype.num_tsarray, dtype.quantity)
    is_classification = data_type in (dtype.categorical, dtype.binary, dtype.cat_tsarray)
    is_multi_ts = tss.is_timeseries and tss.horizon > 1
    has_pretrained_text_enc = any([isinstance(enc, PretrainedLangEncoder)
                                   for enc in encoded_train_data.encoders.values()])

    # raw predictions for validation dataset
    args = {} if not is_classification else {"predict_proba": True}
    normal_predictions = predictor(encoded_val_data, args=PredictionArguments.from_dict(args))
    normal_predictions = normal_predictions.set_index(data.index)

    # ------------------------- #
    # Run analysis blocks, both core and user-defined
    # ------------------------- #
    kwargs = {
        'predictor': predictor,
        'target': target,
        'input_cols': input_cols,
        'dtype_dict': dtype_dict,
        'normal_predictions': normal_predictions,
        'data': data,
        'train_data': train_data,
        'encoded_val_data': encoded_val_data,
        'is_classification': is_classification,
        'is_numerical': is_numerical,
        'is_multi_ts': is_multi_ts,
        'stats_info': stats_info,
        'tss': tss,
        'ts_analysis': ts_analysis,
        'accuracy_functions': accuracy_functions,
        'has_pretrained_text_enc': has_pretrained_text_enc
    }

    for block in analysis_blocks:
        log.info("The block %s is now running its analyze() method", block.__class__.__name__)
        runtime_analyzer = block.analyze(runtime_analyzer, **kwargs)

    # ------------------------- #
    # Populate ModelAnalysis object
    # ------------------------- #
    model_analysis = ModelAnalysis(
        accuracies=runtime_analyzer.get('score_dict', {}),
        accuracy_histogram=runtime_analyzer.get('acc_histogram', {}),
        accuracy_samples=runtime_analyzer.get('acc_samples', {}),
        train_sample_size=len(encoded_train_data),
        test_sample_size=len(encoded_val_data),
        confusion_matrix=runtime_analyzer.get('cm', []),
        column_importances=runtime_analyzer.get('column_importances', {}),
        histograms=stats_info.histograms,
        dtypes=dtype_dict,
        submodel_data=predictor.submodel_data
    )

    return model_analysis, runtime_analyzer



import unittest
import numpy as np
import pandas as pd
from sklearn.metrics import balanced_accuracy_score
from lightwood.api.types import ProblemDefinition
from lightwood.api.high_level import predictor_from_problem, json_ai_from_problem, JsonAI, code_from_json_ai, predictor_from_code  # noqa
np.random.seed(42)


class TestBasic(unittest.TestCase):
    def setup_predictor(self, df, target):
        mask = np.random.rand(len(df)) < 0.8
        train = df[mask]
        test = df[~mask]

        predictor = predictor_from_problem(df, ProblemDefinition.from_dict(
            {'target': target, 'time_aim': 20, 'unbias_target': False}))

        predictor.learn(train)

        if hasattr(predictor, 'ensemble'):
            for i, mixer in enumerate(predictor.ensemble.mixers):

                predictor.ensemble.best_index = i
                predictions = predictor.predict(test)
                assert 'prediction' in predictions.columns
                assert 'confidence' in predictions.columns

                predictions = predictor.predict(test, args={'predict_proba': True})

                for label in df[target].unique():
                    assert f'__mdb_proba_{label}' in predictions.columns
        return predictor

    def test_0_binary(self):
        df = pd.read_csv('tests/data/ionosphere.csv')[:100]
        target = 'target'
        predictor = self.setup_predictor(df, target)
        predictions = predictor.predict(df)
        acc = balanced_accuracy_score(df[target], predictions['prediction'])
        self.assertTrue(acc > 0.5)
        self.assertTrue(all([0 <= p <= 1 for p in predictions['confidence']]))

    def test_1_categorical(self):
        df = pd.read_csv('tests/data/hdi.csv')
        target = 'Development Index'
        predictor = self.setup_predictor(df, target)
        predictions = predictor.predict(df)

        self.assertTrue(balanced_accuracy_score(df[target].astype(int), predictions['prediction'].astype(int)) > 0.9)
        self.assertTrue(all([0 <= p <= 1 for p in predictions['confidence']]))

        # test predict all mixers with some data
        predictions = predictor.predict(df[:10], args={'all_mixers': True})
        assert '__mdb_mixer_Neural' in predictions.columns

        # predict single sample
        predictor.predict(df.iloc[[0]])

    def test_2_binary_no_analysis(self):
        df = pd.read_csv('tests/data/ionosphere.csv')[:100]
        mask = np.random.rand(len(df)) < 0.8
        train = df[mask]
        test = df[~mask]
        predictor = predictor_from_problem(df, ProblemDefinition.from_dict(
            {'target': 'target', 'time_aim': 20, 'use_default_analysis': False}))
        predictor.learn(train)
        predictions = predictor.predict(test)
        self.assertTrue(balanced_accuracy_score(test['target'], predictions['prediction']) > 0.5)
        self.assertTrue('confidence' not in predictions.columns)

    def test_3_test_tempscale_analysis(self):
        # Create base json ai
        df = pd.read_csv('tests/data/hdi.csv').iloc[0:100]
        pdef = ProblemDefinition.from_dict({'target': 'Development Index',
                                            'time_aim': 20,
                                            'use_default_analysis': False
                                            })
        json_ai = json_ai_from_problem(df, pdef)

        # modify it
        json_ai_dump = json_ai.to_dict()
        json_ai_dump['analysis_blocks'] = [{'module': 'TempScaler', 'args': {}}, {'module': 'ConfStats', 'args': {}}]
        json_ai = JsonAI.from_dict(json_ai_dump)

        # create a predictor from it
        code = code_from_json_ai(json_ai)
        predictor = predictor_from_code(code)
        predictor.learn(df)
        row_insights = predictor.predict(df)
        assert 'confidence' in row_insights.columns



from typing import List, Optional, Dict

import pandas as pd
import numpy as np

from lightwood.mixer.base import BaseMixer
from lightwood.ensemble.base import BaseEnsemble
from lightwood.api.types import PredictionArguments
from lightwood.data.encoded_ds import EncodedDs
from lightwood import dtype
from lightwood.helpers.general import evaluate_accuracy
from lightwood.helpers.numeric import is_nan_numeric
from lightwood.helpers.log import log


class ModeEnsemble(BaseEnsemble):
    mixer_scores: Dict[str, float]

    def __init__(self, target, mixers: List[BaseMixer], data: EncodedDs, dtype_dict: dict,
                 accuracy_functions, args: PredictionArguments, ts_analysis: Optional[dict] = None) -> None:
        super().__init__(target, mixers, data)
        self.mixer_scores = {}

        if dtype_dict[target] not in (dtype.binary, dtype.categorical, dtype.tags):
            raise Exception(
                'This ensemble can only be used in classification problems! ' +
                f'Got target dtype {dtype_dict[target]} instead!')

        for _, mixer in enumerate(mixers):
            score_dict = evaluate_accuracy(
                data.data_frame,
                mixer(data, args)['prediction'],
                target,
                accuracy_functions,
                ts_analysis=ts_analysis
            )
            avg_score = np.mean(list(score_dict.values()))
            log.info(f'Mixer: {type(mixer).__name__} got accuracy: {avg_score}')

            if is_nan_numeric(avg_score):
                avg_score = -pow(2, 63)
                log.warning(f'Change the accuracy of mixer {type(mixer).__name__} to valid value: {avg_score}')

            self.mixer_scores[f'__mdb_mixer_{type(mixer).__name__}'] = avg_score

    def _pick_mode_highest_score(self, prediction: pd.Series):
        """If the predictions are unimodal, return the mode. If there are multiple modes, return the mode whose voting
        mixers have the highest score."""
        prediction_counts = prediction.value_counts()

        # If there is a clear winner, i.e. only one prediction
        if len(prediction_counts) == 1:
            return prediction_counts.index[0]

        counts = prediction_counts.values  # how many times all predictions have appeared
        max_count = np.max(counts)  # how many times the most frequent predictions have apppeared

        # most frequent predictions and how many times they appeared
        modes = prediction_counts[prediction_counts == max_count]

        modes_predictions = modes.index  # most frequent predictions

        # For each mode, get the sum of the scores of the predictors who voted for it
        modes_predictions_scores = {}
        for mode_prediction in modes_predictions:
            voting_mixers_name = prediction[prediction == mode_prediction].index.tolist()
            modes_predictions_scores[mode_prediction] = np.sum(
                [self.mixer_scores[mixer_name] for mixer_name in voting_mixers_name])

        # Return the mode with the maximum sum of accuracies
        return max(modes_predictions_scores, key=modes_predictions_scores.get)

    def __call__(self, ds: EncodedDs, args: PredictionArguments) -> pd.DataFrame:
        predictions_df = pd.DataFrame()
        for mixer in self.mixers:
            predictions_df[f'__mdb_mixer_{type(mixer).__name__}'] = mixer(ds, args=args)['prediction']

        mode_df = predictions_df.apply(func=self._pick_mode_highest_score, axis='columns')

        return pd.DataFrame(mode_df, columns=['prediction'])



import importlib
from typing import List, Dict, Optional
import numpy as np
import pandas as pd
from sklearn.metrics import r2_score, f1_score, mean_absolute_error, balanced_accuracy_score
from lightwood.helpers.numeric import is_nan_numeric


def get_group_matches(data, combination):
    """Given a grouped-by combination, return rows of the data that match belong to it. Params:
    data: dict with data to filter and group-by columns info.
    combination: tuple with values to filter by
    return: indexes for rows to normalize, data to normalize
    """
    keys = data['group_info'].keys()  # which column does each combination value belong to

    if isinstance(data['data'], pd.Series):
        data['data'] = np.vstack(data['data'])
    if isinstance(data['data'], np.ndarray) and len(data['data'].shape) < 2:
        data['data'] = np.expand_dims(data['data'], axis=1)

    if combination == '__default':
        idxs = range(len(data['data']))
        return [idxs, np.array(data['data'])[idxs, :]]  # return all data
    else:
        all_sets = []
        for val, key in zip(combination, keys):
            all_sets.append(set([i for i, elt in enumerate(data['group_info'][key]) if elt == val]))
        if all_sets:
            idxs = sorted(list(set.intersection(*all_sets)))
            return idxs, np.array(data['data'])[idxs, :]

        else:
            return [], np.array([])


# ------------------------- #
# Accuracy metrics
# ------------------------- #
def evaluate_accuracy(data: pd.DataFrame,
                      predictions: pd.Series,
                      target: str,
                      accuracy_functions: List[str],
                      ts_analysis: Optional[dict] = {}) -> Dict[str, float]:
    """
    Dispatcher for accuracy evaluation.
    
    :param data: original dataframe.
    :param predictions: output of a lightwood predictor for the input `data`.
    :param target: target column name.
    :param accuracy_functions: list of accuracy function names. Support currently exists for `scikit-learn`'s `metrics` module, plus any custom methods that Lightwood exposes.
    :param ts_analysis: `lightwood.data.timeseries_analyzer` output, used to compute time series task accuracy.
    :return: accuracy metric for a dataset and predictions.
    """  # noqa
    score_dict = {}

    for accuracy_function_str in accuracy_functions:
        if 'array_accuracy' in accuracy_function_str:
            if ts_analysis is None or not ts_analysis['tss'].is_timeseries:
                # normal array, needs to be expanded
                cols = [target]
                true_values = data[cols].apply(lambda x: pd.Series(x[target]), axis=1)
            else:
                horizon = 1 if not isinstance(predictions.iloc[0], list) else len(predictions.iloc[0])
                gby = ts_analysis.get('tss', {}).group_by if ts_analysis.get('tss', {}).group_by else []
                cols = [target] + [f'{target}_timestep_{i}' for i in range(1, horizon)] + gby
                true_values = data[cols]
            predictions = predictions.apply(pd.Series)  # split array cells into columns

            if accuracy_function_str == 'evaluate_array_accuracy':
                acc_fn = evaluate_array_accuracy
            elif accuracy_function_str == 'evaluate_num_array_accuracy':
                acc_fn = evaluate_num_array_accuracy
            elif accuracy_function_str == 'evaluate_cat_array_accuracy':
                acc_fn = evaluate_cat_array_accuracy
            else:
                acc_fn = bounded_evaluate_array_accuracy
            score_dict[accuracy_function_str] = acc_fn(true_values,
                                                       predictions,
                                                       data=data[cols],
                                                       ts_analysis=ts_analysis)
        else:
            true_values = data[target].tolist()
            if hasattr(importlib.import_module('lightwood.helpers.accuracy'), accuracy_function_str):
                accuracy_function = getattr(importlib.import_module('lightwood.helpers.accuracy'),
                                            accuracy_function_str)
            else:
                accuracy_function = getattr(importlib.import_module('sklearn.metrics'), accuracy_function_str)
            score_dict[accuracy_function_str] = accuracy_function(list(true_values), list(predictions))

    return score_dict


def evaluate_regression_accuracy(
        true_values,
        predictions,
        **kwargs
):
    """
    Evaluates accuracy for regression tasks.
    If predictions have a lower and upper bound, then `within-bound` accuracy is computed: whether the ground truth value falls within the predicted region.
    If not, then a (positive bounded) R2 score is returned instead.
    
    :return: accuracy score as defined above. 
    """  # noqa
    if 'lower' and 'upper' in predictions:
        Y = np.array(true_values).astype(float)
        within = ((Y >= predictions['lower']) & (Y <= predictions['upper']))
        return sum(within) / len(within)
    else:
        r2 = r2_score(true_values, predictions['prediction'])
        return max(r2, 0)


def evaluate_multilabel_accuracy(true_values, predictions, **kwargs):
    """
    Evaluates accuracy for multilabel/tag prediction.

    :return: weighted f1 score of predictions and ground truths.
    """
    pred_values = predictions['prediction']
    return f1_score(true_values, pred_values, average='weighted')


def evaluate_num_array_accuracy(
        true_values: pd.Series,
        predictions: pd.Series,
        **kwargs
) -> float:
    """
    Evaluate accuracy in numerical time series forecasting tasks.
    Defaults to mean absolute scaled error (MASE) if in-sample residuals are available.
    If this is not the case, R2 score is computed instead.

    Scores are computed for each timestep (as determined by `timeseries_settings.horizon`),
    and the final accuracy is the reciprocal of the average score through all timesteps.
    """
    ts_analysis = kwargs.get('ts_analysis', {})
    if not ts_analysis:
        naive_errors = None
    else:
        naive_errors = ts_analysis.get('ts_naive_mae', {})
        wrapped_data = {
            'data': kwargs['data'].reset_index(drop=True),
            'group_info': {gcol: kwargs['data'][gcol].tolist()
                           for gcol in ts_analysis['tss'].group_by} if ts_analysis['tss'].group_by else {}
        }

        if ts_analysis['tss'].group_by:
            [true_values.pop(gby_col) for gby_col in ts_analysis['tss'].group_by]

    true_values = np.array(true_values)
    predictions = np.array(predictions)

    if not naive_errors:
        # use mean R2 method if naive errors are not available
        nan_mask = (~np.isnan(true_values)).astype(int)
        predictions *= nan_mask
        true_values = np.nan_to_num(true_values, 0.0)
        return evaluate_array_accuracy(true_values, predictions, ts_analysis=ts_analysis)

    mases = []
    for group in ts_analysis['group_combinations']:
        g_idxs, _ = get_group_matches(wrapped_data, group)

        # only evaluate populated groups
        if g_idxs:
            trues = true_values[g_idxs]
            preds = predictions[g_idxs]

            # add MASE score for each group (__default only considered if the task is non-grouped)
            if len(ts_analysis['group_combinations']) == 1 or group != '__default':
                mases.append(mase(trues, preds, ts_analysis['ts_naive_mae'][group], ts_analysis['tss'].horizon))

    return 1 / max(np.average(mases), 1e-4)  # reciprocal to respect "larger -> better" convention


def evaluate_array_accuracy(
        true_values: np.ndarray,
        predictions: np.ndarray,
        **kwargs
) -> float:
    """
    Default time series forecasting accuracy method.
    Returns mean score over all timesteps in the forecasting horizon, as determined by the `base_acc_fn` (R2 score by default).
    """  # noqa
    base_acc_fn = kwargs.get('base_acc_fn', lambda t, p: max(0, r2_score(t, p)))

    fh = true_values.shape[1]
    ts_analysis = kwargs.get('ts_analysis', None)

    if ts_analysis and ts_analysis.get('tss', False) and not kwargs['ts_analysis']['tss'].eval_cold_start:
        # only evaluate accuracy for rows with complete historical context
        true_values = true_values[kwargs['ts_analysis']['tss'].window:]
        predictions = predictions[kwargs['ts_analysis']['tss'].window:]

    aggregate = 0.0
    for i in range(fh):
        aggregate += base_acc_fn([t[i] for t in true_values], [p[i] for p in predictions])

    return aggregate / fh


def evaluate_cat_array_accuracy(
        true_values: pd.Series,
        predictions: pd.Series,
        **kwargs
) -> float:
    """
    Evaluate accuracy in categorical time series forecasting tasks.

    Balanced accuracy is computed for each timestep (as determined by `timeseries_settings.horizon`),
    and the final accuracy is the reciprocal of the average score through all timesteps.
    """
    ts_analysis = kwargs.get('ts_analysis', {})

    if ts_analysis and ts_analysis['tss'].group_by:
        [true_values.pop(gby_col) for gby_col in ts_analysis['tss'].group_by]

    true_values = np.array(true_values)
    predictions = np.array(predictions)

    return evaluate_array_accuracy(true_values,
                                   predictions,
                                   ts_analysis=ts_analysis,
                                   base_acc_fn=balanced_accuracy_score)


def bounded_evaluate_array_accuracy(
        true_values: pd.Series,
        predictions: pd.Series,
        **kwargs
) -> float:
    """
    The normal MASE accuracy inside ``evaluate_array_accuracy`` has a break point of 1.0: smaller values mean a naive forecast is better, and bigger values imply the forecast is better than a naive one. It is upper-bounded by 1e4.

    This 0-1 bounded MASE variant scores the 1.0 breakpoint to be equal to 0.5.
    For worse-than-naive, it scales linearly (with a factor).
    For better-than-naive, we fix 10 as 0.99, and scaled-logarithms (with 10 and 1e4 cutoffs as respective bases) are used to squash all remaining preimages to values between 0.5 and 1.0.
    """  # noqa
    result = evaluate_array_accuracy(np.array(true_values),
                                     np.array(predictions),
                                     **kwargs)
    if 10 < result <= 1e4:
        step_base = 0.99
        return step_base + (np.log(result) / np.log(1e4)) * (1 - step_base)
    elif 1 <= result <= 10:
        step_base = 0.5
        return step_base + (np.log(result) / np.log(10)) * (0.99 - step_base)
    else:
        return result / 2  # worse than naive


# ------------------------- #
# Helpers
# ------------------------- #
def mase(trues, preds, scale_error, fh):
    """
    Computes mean absolute scaled error.
    The scale corrective factor is the mean in-sample residual from the naive forecasting method.
    """
    if scale_error == 0:
        scale_error = 1  # cover (rare) case where series is constant

    nan_mask = (~np.isnan(trues)).astype(int)
    preds *= nan_mask
    trues = np.nan_to_num(trues, 0.0)

    agg = 0.0
    for i in range(fh):
        true = trues[:, i]
        pred = preds[:, i]
        agg += mean_absolute_error(true, pred)

    return (agg / fh) / scale_error


def is_none(value):
    """
    We use pandas :(
    Pandas has no way to guarantee "stability" for the type of a column, it choses to arbitrarily change it based on the values.
    Pandas also change the values in the columns based on the types.
    Lightwood relies on having ``None`` values for a cells that represent "missing" or "corrupt".
    
    When we assign ``None`` to a cell in a dataframe this might get turned to `nan` or other values, this function checks if a cell is ``None`` or any other values a pd.DataFrame might convert ``None`` to.

    It also check some extra values (like ``''``) that pandas never converts ``None`` to (hopefully). But lightwood would still consider those values "None values", and this will allow for more generic use later.
    """ # noqa
    if value is None:
        return True

    if is_nan_numeric(value):
        return True

    if str(value) == '':
        return True

    if str(value) in ('None', 'nan', 'NaN', 'np.nan'):
        return True

    return False




import unittest
import torch
from typing import List
from lightwood.encoder.identity.identity import IdentityEncoder


class TestIdentityEncoder(unittest.TestCase):
    def test_encode_and_decode(self):
        data = [1, 1.1, 2, -8.6, 0]

        encoder = IdentityEncoder()

        encoded_vals = encoder.encode(data)

        self.assertTrue(isinstance(encoded_vals, torch.Tensor))

        for i in range(len(encoded_vals)):
            self.assertTrue(abs(encoded_vals[i].item() - data[i]) <= 1e-5)

        decoded_vals = encoder.decode(encoded_vals)

        self.assertTrue(isinstance(decoded_vals, List))

        for i in range(len(decoded_vals)):
            self.assertTrue(abs(decoded_vals[i] - data[i]) <= 1e-5)



import unittest
import numpy as np
import torch
from lightwood.encoder.numeric import NumericEncoder
from lightwood.encoder.numeric import TsNumericEncoder
from lightwood.helpers.general import is_none


def _pollute(array):
    return [
        array + [np.nan],
        array + [np.inf],
        array + [None]
    ]


class TestNumericEncoder(unittest.TestCase):
    def test_encode_and_decode(self):
        data = [1, 1.1, 2, -8.6, None, 0]

        encoder = NumericEncoder()

        encoder.prepare(data)
        encoded_vals = encoder.encode(data)

        self.assertTrue(encoded_vals[1][1] > 0)
        self.assertTrue(encoded_vals[2][1] > 0)
        self.assertTrue(encoded_vals[3][1] > 0)
        for i in range(0, 3):
            self.assertTrue(encoded_vals[i][2] == 0)
        self.assertTrue(encoded_vals[3][2] == 1)
        self.assertTrue(encoded_vals[4][3] == 0)

        decoded_vals = encoder.decode(encoded_vals)

        for i in range(len(encoded_vals)):
            if decoded_vals[i] is None:
                self.assertTrue(decoded_vals[i] == data[i])
            else:
                np.testing.assert_almost_equal(round(decoded_vals[i], 10), round(data[i], 10))

    def test_positive_domain(self):
        data = [-1, -2, -100, 5, 10, 15]
        for encoder in [NumericEncoder(), TsNumericEncoder()]:
            encoder.is_target = True        # only affects target values
            encoder.positive_domain = True
            encoder.prepare(data)
            decoded_vals = encoder.decode(encoder.encode(data))

            for val in decoded_vals:
                self.assertTrue(val >= 0)

    def test_log_overflow_and_none(self):
        data = list(range(-2000, 2000, 66))
        encoder = NumericEncoder()

        encoder.is_target = True
        encoder.positive_domain = True
        encoder.decode_log = True
        encoder.prepare(data)
        encoder.decode(encoder.encode(data))

        for i in range(0, 70, 10):
            encoder.decode([[0, pow(2, i), 0]])

    def test_nan_encoding(self):
        # Generate some numbers
        data = list(range(-50, 50, 2))

        # Add invalid values to the data
        invalid_data = _pollute(data)

        # Prepare with the correct data and decode invalid data
        encoder = NumericEncoder()
        encoder.prepare(data)
        for array in invalid_data:
            # Make sure the encoding has no nans or infs
            encoded_repr = encoder.encode(array)
            assert not torch.isnan(encoded_repr).any()
            assert not torch.isinf(encoded_repr).any()

            # Make sure the invalid value is decoded as `None` and the rest as numbers
            decoded_repr = encoder.decode(encoded_repr)
            for x in decoded_repr[:-1]:
                assert not is_none(x)
            assert decoded_repr[-1] is None

        # Prepare with the invalid data and decode the valid data
        for array in invalid_data:
            encoder = NumericEncoder()
            encoder.prepare(array)

            # Make sure the encoding has no nans or infs
            encoded_repr = encoder.encode(data)
            assert not torch.isnan(encoded_repr).any()
            assert not torch.isinf(encoded_repr).any()

            # Make sure the invalid value is decoded as `None` and the rest as numbers
            decoded_repr = encoder.decode(encoded_repr)
            for x in decoded_repr:
                assert not is_none(x)

        # Prepare with the invalid data and decode invalid data
        for array in invalid_data:
            encoder = NumericEncoder()
            encoder.prepare(array)
            # Make sure the encoding has no nans or infs
            encoded_repr = encoder.encode(array)
            assert not torch.isnan(encoded_repr).any()
            assert not torch.isinf(encoded_repr).any()

            # Make sure the invalid value is decoded as `None` and the rest as numbers
            decoded_repr = encoder.decode(encoded_repr)
            for x in decoded_repr[:-1]:
                assert not is_none(x)
            assert decoded_repr[-1] is None



import unittest
import numpy as np
import pandas as pd
from lightwood.api.types import ProblemDefinition
from lightwood.api.high_level import json_ai_from_problem, code_from_json_ai, predictor_from_code, predictor_from_problem  # noqa

np.random.seed(0)


class TestArrayTarget(unittest.TestCase):
    def _test_array(self, df):
        train_idxs = np.random.rand(len(df)) < 0.8
        train = df[train_idxs]
        test = df[~train_idxs]

        predictor = predictor_from_problem(df,
                                           ProblemDefinition.from_dict({'target': 'output',
                                                                        'time_aim': 80,
                                                                        }))
        predictor.learn(train)
        predictor.predict(test)

    def test_0_num_array(self):
        """ Tests numerical array input and output. """
        # task: learn to emit next `arr_len` numbers given any numerical array input of length `arr_len`
        df = pd.DataFrame()
        arr_len = 4
        df['input'] = [[row + i for i in range(arr_len)] for row in range(200)]
        df['output'] = [[row + i + arr_len for i in range(arr_len)] for row in range(200)]
        self._test_array(df)

    def test_1_cat_array(self):
        """ Tests categorical array input and output. """
        # task: learn to reverse the `arr_len`-length input array.
        df = pd.DataFrame()
        arr_len = 4
        df['input'] = [[chr(65 + i + row % 4) for i in range(arr_len)] for row in range(200)]
        df['output'] = [[chr(65 + i + row % 4) for i in range(arr_len)][::-1] for row in range(200)]
        self._test_array(df)



from typing import Iterable


def is_nan_numeric(value: object) -> bool:
    """
    Determines if **value** might be `nan` or `inf` or some other numeric value (i.e. which can be cast as `float`) that is not actually a number.
    """  # noqa

    try:
        value = str(value)
        value = float(value)
    except Exception:
        return False

    try:
        if isinstance(value, float):
            a = int(value) # noqa
        isnan = False
    except Exception:
        isnan = True
    return isnan


def filter_nan_and_none(series: Iterable) -> list:
    return [x for x in series if not is_nan_numeric(x) and x is not None]



import unittest
import torch
from torch import Tensor
import numpy as np
from lightwood.encoder.categorical.binary import (
    BinaryEncoder,
)

from lightwood.helpers.constants import _UNCOMMON_WORD


class TestBinary(unittest.TestCase):
    """ Test the OHE vector encoder """

    def test_encode_decode_with_binary(self):
        """
        Test binary example case.
        """
        # Generate categories with None
        data = ["apple", "apple", "orange", None, "apple", "orange"]

        # Generate test category with unseen examples (banana, None)
        test_data = ["apple", "banana", "orange", None, "apple"]

        # Ground-truth answer
        ytest = ["apple", _UNCOMMON_WORD, "orange", _UNCOMMON_WORD, "apple"]

        enc = BinaryEncoder()
        enc.prepare(data)

        enc_data = enc.encode(data)
        dec_data = enc.decode(enc.encode(test_data))

        self.assertTrue(
            (
                enc_data
                == Tensor(
                    [
                        [
                            1.0,
                            0.0,
                        ],  # category 1
                        [
                            1.0,
                            0.0,
                        ],  # category 1
                        [
                            0.0,
                            1.0,
                        ],  # category 2
                        [
                            0.0,
                            0.0,
                        ],  # None
                        [
                            1.0,
                            0.0,
                        ],  # category 1
                        [
                            0.0,
                            1.0,
                        ],  # category 2
                    ]
                )
            ).all()
        )

        for i in range(len(ytest)):
            self.assertTrue(dec_data[i] == ytest[i])

    def test_check_only_binary(self):
        """ Ensure binary strictly enforces binary typing """
        data = ["apple", "apple", "orange", "banana", "apple", "orange"]

        enc = BinaryEncoder()
        self.assertRaises(ValueError, enc.prepare, data)

    def test_check_probabilities(self):
        """
        Check whether decode_probabilities returns valid scaled-to-1 terms
        """
        # Generate random list of T/F
        np.random.seed(1)
        data = np.random.rand(10) > 0.5

        enc = BinaryEncoder()
        enc.prepare(data)

        # Make data to represent random weights that do not sum to 1
        torch.manual_seed(1)
        wt_vec = torch.rand(size=(len(data), len(enc.map))) * 10

        _, probs, _ = enc.decode_probabilities(wt_vec)
        self.assertTrue(np.all([np.isclose(sum(i), 1) for i in probs]))

    def test_target_distro_scaled_to_1(self):
        """
        Check whether target distribution passed and handled properly
        """
        data = ["apple", "apple", "orange", "apple", "apple", "orange"]

        # Scaled weights (sum to 1)
        tweights = {"apple": 4 / 6, "orange": 2 / 6}

        enc = BinaryEncoder(is_target=True, target_weights=tweights)
        enc.prepare(data)

        # Get the ground-truth weights
        iweights = torch.ones(size=(len(tweights),))
        for key, value in tweights.items():  # accounts for order
            iweights[enc.map[key]] = value

        # Check inverse weights correct
        self.assertTrue(np.all(((enc.index_weights - iweights) == 0).tolist()))

    def test_distro_nonzeroweights(self):
        """
        Tests if target weights do not sum to 1 properly handled.

        This handles cases where people may choose 1/class_size for weights
        """
        data = ["apple", "apple", "orange", "apple", "apple", "orange"]
        tweights = {"apple": 100, "orange": 5000}

        enc = BinaryEncoder(is_target=True, target_weights=tweights)
        enc.prepare(data)

        # Get the ground-truth weights
        iweights = torch.ones(size=(len(tweights),))
        for key, value in tweights.items():  # accounts for order
            iweights[enc.map[key]] = value

        self.assertTrue(np.all(((enc.index_weights - iweights) == 0).tolist()))

    def test_distro_zero(self):
        """
        Tests edge cause where target weights have a 0 weight which is unacceptable for downstream processing (inverse weights will 1/0)
        """ # noqa
        data = ["apple", "apple", "orange", "apple", "apple", "orange"]

        # Arbitrary weights (ex: number of examples)
        tweights = {"apple": 100, "orange": 0}

        enc = BinaryEncoder(is_target=True, target_weights=tweights)

        # Check if 0-weight class is rejected
        self.assertRaises(ValueError, enc.prepare, data)


if __name__ == "__main__":
    unittest.main()



import re
import datetime
from copy import deepcopy
from dateutil.parser import parse as parse_dt
from typing import Dict, List, Optional, Tuple, Callable, Union

import numpy as np
import pandas as pd

from lightwood.api.dtype import dtype
from lightwood.helpers import text
from lightwood.helpers.log import log
from lightwood.helpers.imputers import BaseImputer
from lightwood.api.types import TimeseriesSettings
from lightwood.helpers.numeric import is_nan_numeric


def cleaner(
    data: pd.DataFrame,
    dtype_dict: Dict[str, str],
    pct_invalid: float,
    identifiers: Dict[str, str],
    target: str,
    mode: str,
    timeseries_settings: TimeseriesSettings,
    anomaly_detection: bool,
    imputers: Dict[str, BaseImputer] = {},
    custom_cleaning_functions: Dict[str, str] = {}
) -> pd.DataFrame:
    """
    The cleaner is a function which takes in the raw data, plus additional information about it's types and about the problem. Based on this it generates a "clean" representation of the data, where each column has an ideal standardized type and all malformed or otherwise missing or invalid elements are turned into ``None``. Optionally, these ``None`` values can be replaced with imputers.

    :param data: The raw data
    :param dtype_dict: Type information for each column
    :param pct_invalid: How much of each column can be invalid
    :param identifiers: A dict containing all identifier typed columns
    :param target: The target columns
    :param mode: Can be "predict" or "train"
    :param imputers: The key corresponds to the single input column that will be imputed by the object. Refer to the imputer documentation for more details.
    :param timeseries_settings: Timeseries related settings, only relevant for timeseries predictors, otherwise can be the default object
    :param anomaly_detection: Are we detecting anomalies with this predictor?

    :returns: The cleaned data
    """ # noqa

    data = _remove_columns(data, identifiers, target, mode, timeseries_settings,
                           anomaly_detection, dtype_dict)

    data['__mdb_original_index'] = np.arange(len(data))

    for col in _get_columns_to_clean(data, dtype_dict, mode, target):

        # Get and apply a cleaning function for each data type
        # If you want to customize the cleaner, it's likely you can to modify ``get_cleaning_func``
        data[col] = data[col].apply(get_cleaning_func(dtype_dict[col], custom_cleaning_functions))

    if timeseries_settings.is_timeseries:
        data = clean_timeseries(data, timeseries_settings)

    for col, imputer in imputers.items():
        if col in data.columns:
            cols = [col] + [col for col in imputer.dependencies]
            data[col] = imputer.impute(data[cols])

    return data


def _check_if_invalid(new_data: pd.Series, pct_invalid: float, col_name: str):
    """
    Checks how many invalid data points there are. Invalid data points are flagged as "Nones" from the cleaning processs (see data/cleaner.py for default).
    If there are too many invalid data points (specified by `pct_invalid`), then an error message will pop up. This is used as a safeguard for very messy data.

    :param new_data: data to check for invalid values.
    :param pct_invalid: maximum percentage of invalid values. If this threshold is surpassed, an exception is raised.
    :param col_name: name of the column to analyze.

    """  # noqa

    chk_invalid = (
        100
        * (len(new_data) - len([x for x in new_data if x is not None]))
        / len(new_data)
    )

    if chk_invalid > pct_invalid:
        err = f'Too many ({chk_invalid}%) invalid values in column {col_name}nam'
        log.error(err)
        raise Exception(err)


def get_cleaning_func(data_dtype: dtype, custom_cleaning_functions: Dict[str, str]) -> Callable:
    """
    For the provided data type, provide the appropriate cleaning function. Below are the defaults, users can either override this function OR impose a custom block.

    :param data_dtype: The data-type (inferred from a column) as prescribed from ``api.dtype``

    :returns: The appropriate function that will pre-process (clean) data of specified dtype.
    """ # noqa
    if data_dtype in custom_cleaning_functions:
        clean_func = eval(custom_cleaning_functions[data_dtype])

    elif data_dtype in (dtype.date, dtype.datetime):
        clean_func = _standardize_datetime

    elif data_dtype in (dtype.float, dtype.num_tsarray):
        clean_func = _clean_float

    elif data_dtype in (dtype.integer):
        clean_func = _clean_int

    elif data_dtype in (dtype.num_array):
        clean_func = _standardize_num_array

    elif data_dtype in (dtype.cat_array):
        clean_func = _standardize_cat_array

    elif data_dtype in (dtype.tags):
        clean_func = _tags_to_tuples

    elif data_dtype in (dtype.quantity):
        clean_func = _clean_quantity

    elif data_dtype in (
        dtype.short_text,
        dtype.rich_text,
        dtype.categorical,
        dtype.binary,
        dtype.audio,
        dtype.image,
        dtype.video,
        dtype.cat_tsarray
    ):
        clean_func = _clean_text

    else:
        raise ValueError(f"{data_dtype} is not supported. Check lightwood.api.dtype")

    return clean_func


# ------------------------- #
# Temporal Cleaning
# ------------------------- #


def _standardize_datetime(element: object) -> Optional[float]:
    """
    Parses an expected date-time element. Intakes an element that can in theory be anything.
    """
    try:
        date = parse_dt(str(element))
    except Exception:
        try:
            date = datetime.datetime.utcfromtimestamp(element)
        except Exception:
            return None

    return date.timestamp()


# ------------------------- #
# Tags/Sequences
# ------------------------- #

# TODO Make it split on something other than commas
def _tags_to_tuples(tags_str: str) -> Tuple[str]:
    """
    Converts comma-separated values into a tuple to preserve a sequence/array.

    Ex:
    >> x = 'apples, oranges, bananas'
    >> _tags_to_tuples(x)
    >> ('apples', 'oranges', 'bananas')
    """
    try:
        return tuple([x.strip() for x in tags_str.split(",")])
    except Exception:
        return tuple()


def _standardize_num_array(element: object) -> Optional[Union[List[float], float]]:
    """
    Given an array of numbers in the form ``[1, 2, 3, 4]``, converts into a numerical sequence.

    :param element: An array-like element in a sequence
    :returns: standardized array OR scalar number IF edge case

    Ex of edge case:
    >> element = [1]
    >> _standardize_num_array(element)
    >> 1
    """
    try:
        element = str(element)
        element = element.rstrip("]").lstrip("[")
        element = element.rstrip(" ").lstrip(" ")
        element = element.replace(", ", " ").replace(",", " ")
        # Handles cases where arrays are numbers
        if " " not in element:
            element = _clean_float(element)
        else:
            element = [float(x) for x in element.split(" ")]
    except Exception:
        pass

    return element


def _standardize_cat_array(element: List) -> Optional[List[str]]:
    """
    Given an array, replace non-string with string casted (or, failing that, None) tokens. None values are kept.

    :param element: An array-like element in a sequence
    :returns: standardized array OR label IF edge case

    Ex of edge case:
    >> element = ['a']
    >> _standardize_cat_array(element)
    >> 'a'
    """
    if len(element) == 1:
        return element[0]
    else:
        new_element = []
        for sub_elt in element:
            try:
                new_sub_elt = str(sub_elt) if sub_elt else None
                new_element.append(new_sub_elt)
            except TypeError:
                new_element.append(None)
        element = new_element

    return element


# ------------------------- #
# Integers/Floats/Quantities
# ------------------------- #

def _clean_float(element: object) -> Optional[float]:
    """
    Given an element, converts it into float numeric format. If element is NaN, or inf, then returns None.
    """
    try:
        cleaned_float = text.clean_float(element)
        if is_nan_numeric(cleaned_float):
            return None
        return cleaned_float
    except Exception:
        return None


def _clean_int(element: object) -> Optional[int]:
    """
    Given an element, converts it into integer numeric format. If element is NaN, or inf, then returns None.
    """
    element = _clean_float(element)
    if element is not None:
        element = int(element)
    return element


def _clean_quantity(element: object) -> Optional[float]:
    """
    Given a quantity, clean and convert it into float numeric format. If element is NaN, or inf, then returns None.
    """
    no_symbols = re.sub("[^0-9.,]", "", str(element)).replace(",", ".")
    no_symbols = '0' if no_symbols == '' else no_symbols
    element = float(no_symbols)
    return _clean_float(element)


# ------------------------- #
# Text
# ------------------------- #
def _clean_text(element: object) -> Union[str, None]:
    if isinstance(element, str):
        return element
    elif element is None or element != element:
        return None
    else:
        return str(element)


# ------------------------- #
# Other helpers
# ------------------------- #
def _rm_rows_w_empty_targets(df: pd.DataFrame, target: str) -> pd.DataFrame:
    """
    Drop any rows that have targets as unknown. Targets are necessary to train.

    :param df: The input dataframe including the target value
    :param target: the column name that is the output target variable

    :returns: Data with any target smissing
    """
    # Compare length before/after
    len_before = len(df)

    # Use Pandas ```dropna``` to omit any rows with missing values for targets; these cannot be trained
    df = df.dropna(subset=[target])

    # Compare length with after
    len_after = len(df)
    nr_removed = len_before - len_after

    if nr_removed != 0:
        log.warning(
            f"Removed {nr_removed} rows because target was missing. Training on these rows is not possible."
        )  # noqa

    return df


def _remove_columns(data: pd.DataFrame, identifiers: Dict[str, object], target: str,
                    mode: str, timeseries_settings: TimeseriesSettings, anomaly_detection: bool,
                    dtype_dict: Dict[str, dtype]) -> pd.DataFrame:
    """
    Drop columns we don't want to use in order to train or predict

    :param data: The raw data
    :param dtype_dict: Type information for each column
    :param identifiers: A dict containing all identifier typed columns
    :param target: The target columns
    :param mode: Can be "predict" or "train"
    :param timeseries_settings: Timeseries related settings, only relevant for timeseries predictors, otherwise can be the default object
    :param anomaly_detection: Are we detecting anomalies with this predictor?

    :returns: A (new) dataframe without the dropped columns
    """ # noqa
    data = deepcopy(data)
    to_drop = [*[x for x in identifiers.keys() if x != target],
               *[x for x in data.columns if x in dtype_dict and dtype_dict[x] == dtype.invalid]]

    exceptions = ["__mdb_forecast_offset"]
    if timeseries_settings.group_by is not None:
        exceptions += timeseries_settings.group_by

    to_drop = [x for x in to_drop if x in data.columns and x not in exceptions]

    if to_drop:
        log.info(f'Dropping features: {to_drop}')
        data = data.drop(columns=to_drop)

    if mode == "train":
        data = _rm_rows_w_empty_targets(data, target)
    if mode == "predict":
        if (
            target in data.columns
            and (not timeseries_settings.is_timeseries or not timeseries_settings.use_previous_target)
            and not anomaly_detection
        ):
            data = data.drop(columns=[target])

    # Drop extra columns
    for name in list(data.columns):
        if name not in dtype_dict and name not in exceptions:
            data = data.drop(columns=[name])
    return data


def _get_columns_to_clean(data: pd.DataFrame, dtype_dict: Dict[str, dtype], mode: str, target: str) -> List[str]:
    """
    :param data: The raw data
    :param dtype_dict: Type information for each column
    :param target: The target columns
    :param mode: Can be "predict" or "train"

    :returns: A list of columns that we want to clean
    """ # noqa

    cleanable_columns = []
    for name, _ in dtype_dict.items():
        if mode == "predict":
            if name == target:
                continue
        if name in data.columns:
            cleanable_columns.append(name)
    return cleanable_columns


def clean_timeseries(df: pd.DataFrame, tss: TimeseriesSettings) -> pd.DataFrame:
    """
    All timeseries-specific cleaning logic goes here. Currently:

    1) Any row with `nan`-valued order-by measurements is dropped.

    :param df: data.
    :param tss: timeseries settings
    :return: cleaned data.
    """
    invalid_rows = []

    for idx, row in df.iterrows():
        if pd.isna(row[tss.order_by[0]]):
            invalid_rows.append(idx)

    df = df.drop(invalid_rows)

    return df



import inspect
import importlib
from datetime import datetime
from itertools import product
from typing import Dict, Union

import optuna
import numpy as np
import pandas as pd
from sktime.forecasting.trend import PolynomialTrendForecaster
from sktime.forecasting.compose import TransformedTargetForecaster
from sktime.transformations.series.detrend import ConditionalDeseasonalizer
from sktime.transformations.series.detrend import Detrender
from sktime.forecasting.base import ForecastingHorizon, BaseForecaster
from sktime.performance_metrics.forecasting import MeanAbsolutePercentageError
try:
    from sktime.forecasting.statsforecast import StatsForecastAutoARIMA as AutoARIMA
except ModuleNotFoundError:
    from sktime.forecasting.arima import AutoARIMA

from lightwood.helpers.log import log
from lightwood.mixer.base import BaseMixer
from lightwood.api.types import PredictionArguments
from lightwood.helpers.general import get_group_matches
from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs


class SkTime(BaseMixer):
    forecaster: str
    horizon: int
    target: str
    supports_proba: bool
    model_path: str
    hyperparam_search: bool

    def __init__(
            self,
            stop_after: float,
            target: str,
            dtype_dict: Dict[str, str],
            horizon: int,
            ts_analysis: Dict,
            model_path: str = 'arima.AutoARIMA',
            auto_size: bool = True,
            hyperparam_search: bool = False,
            target_transforms: Dict[str, Union[int, str]] = {}):
        """
        This mixer is a wrapper around the popular time series library sktime. It exhibits different behavior compared
        to other forecasting mixers, as it predicts based on indices in a forecasting horizon that is defined with
        respect to the last seen data point at training time.
        
        Due to this, the mixer tries to "fit_on_all" so that the latest point in the validation split marks the 
        difference between training data and where forecasts will start. In practice, you need to specify how much 
        time has passed since the aforementioned timestamp for correct forecasts. By default, it is assumed that
         predictions are for the very next timestamp post-training.
        
        If the task has groups (i.e. 'TimeseriesSettings.group_by' is not empty), the mixer will spawn one forecaster 
        object per each different group observed at training time, plus an additional default forecaster fit with all data.
        
        There is an optuna-based automatic hyperparameter search. For now, it considers selecting the forecaster type
        based on the global SMAPE error across all groups.
        
        :param stop_after: time budget in seconds.
        :param target: column to forecast.
        :param dtype_dict: dtypes of all columns in the data.
        :param horizon: length of forecasted horizon.
        :param ts_analysis: dictionary with miscellaneous time series info, as generated by 'lightwood.data.timeseries_analyzer'.
        :param model_path: sktime forecaster to use as underlying model(s). Should be a string with format "$module.$class' where '$module' is inside `sktime.forecasting`. Default is 'arima.AutoARIMA'.
        :param hyperparam_search: bool that indicates whether to perform the hyperparameter tuning or not.
        :param auto_size: whether to filter out old data points if training split is bigger than a certain threshold (defined by the dataset sampling frequency). Enabled by default to avoid long training times in big datasets.
        :param target_transforms: arguments for target transformation. Currently supported format: {'detrender': int, 'deseasonalizer': 'add' | 'mul' }. 'detrender' forces a particular type of polynomial to fit as trend curve for the series, while 'deseasonalizer' specifies additive or multiplicative seasonality decomposition (only applied if a seasonality test is triggered). By default, both are disabled.
        """  # noqa
        super().__init__(stop_after)
        self.stable = True
        self.prepared = False
        self.supports_proba = False
        self.target = target

        self.ts_analysis = ts_analysis
        self.horizon = horizon
        self.grouped_by = ['__default'] if not ts_analysis['tss'].group_by else ts_analysis['tss'].group_by
        self.auto_size = auto_size
        self.cutoff_factor = 4  # times the detected maximum seasonal period
        self.target_transforms = {
            'detrender': 0,          # degree of detrender polynomial (0: disabled)
            'deseasonalizer': ''  # seasonality decomposition: 'add'itive or 'mul'tiplicative (else, disabled)
        }
        self.target_transforms.update(target_transforms)

        # optuna hyperparameter tuning
        self.models = {}
        self.study = None
        self.hyperparam_dict = {}
        self.model_path = model_path
        self.hyperparam_search = hyperparam_search
        self.trial_error_fn = MeanAbsolutePercentageError(symmetric=True)
        self.possible_models = ['ets.AutoETS', 'theta.ThetaForecaster', 'arima.AutoARIMA']
        self.n_trials = len(self.possible_models)
        self.freq = self._get_freq(self.ts_analysis['deltas']['__default'][self.ts_analysis['tss'].order_by[0]])

        # sktime forecast horizon object is made relative to the end of the latest data point seen at training time
        # the default assumption is to forecast the next `self.horizon` after said data point
        self.fh = ForecastingHorizon(np.arange(1, self.horizon + 1), is_relative=True)

    def fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Fits a set of sktime forecasters. The number of models depends on how many groups are observed at training time.

        Forecaster type can be specified by providing the `model_class` argument in `__init__()`. It can also be determined by hyperparameter optimization based on dev data validation error.
        """  # noqa
        log.info('Started fitting sktime forecaster for array prediction')

        if self.hyperparam_search:
            search_space = {'class': self.possible_models}
            self.study = optuna.create_study(direction='minimize', sampler=optuna.samplers.GridSampler(search_space))
            self.study.optimize(lambda trial: self._get_best_model(trial, train_data, dev_data), n_trials=self.n_trials)
            data = ConcatedEncodedDs([train_data, dev_data])
            self._fit(data)

        else:
            data = ConcatedEncodedDs([train_data, dev_data])
            self._fit(data)

    def _fit(self, data):
        """
        Internal method that fits forecasters to a given dataframe.
        """
        df = data.data_frame.sort_values(by=f'__mdb_original_{self.ts_analysis["tss"].order_by[0]}')
        data = {'data': df,
                'group_info': {gcol: df[gcol].tolist()
                               for gcol in self.grouped_by} if self.ts_analysis['tss'].group_by
                else {'': ['__default' for _ in range(df.shape[0])]}}

        if not self.hyperparam_search and not self.study:
            module_name = self.model_path
        else:
            finished_study = sum([int(trial.state.is_finished()) for trial in self.study.trials]) == self.n_trials
            if finished_study:
                log.info(f'Selected best model: {self.study.best_params["class"]}')
                module_name = self.study.best_params['class']
            else:
                module_name = self.hyperparam_dict['class']  # select active optuna choice

        sktime_module = importlib.import_module('.'.join(['sktime', 'forecasting', module_name.split(".")[0]]))
        try:
            model_class = getattr(sktime_module, module_name.split(".")[1])
        except AttributeError:
            model_class = AutoARIMA  # use AutoARIMA when the provided class does not exist

        for group in self.ts_analysis['group_combinations']:
            kwargs = {}
            options = {
                'sp': self.ts_analysis['periods'].get(group, '__default'),  # seasonality period
                'suppress_warnings': True,                                  # ignore warnings if possible
                'error_action': 'raise',                                    # avoids fit() failing silently
            }
            if self.model_path == 'fbprophet.Prophet':
                options['freq'] = self.freq

            for k, v in options.items():
                kwargs = self._add_forecaster_kwarg(model_class, kwargs, k, v)

            model_pipeline = [("forecaster", model_class(**kwargs))]

            trend_degree = self.target_transforms['detrender']
            seasonality_type = self.target_transforms['deseasonalizer']
            if seasonality_type in ('add', 'mul'):
                model_pipeline.insert(0, ("deseasonalizer",
                                          ConditionalDeseasonalizer(
                                              model='additive' if seasonality_type == 'add' else 'multiplicative',
                                              sp=options['sp']
                                          )))
            if trend_degree > 0:
                model_pipeline.insert(0, ("detrender",
                                          Detrender(forecaster=PolynomialTrendForecaster(degree=trend_degree))))

            self.models[group] = TransformedTargetForecaster(model_pipeline)

            oby_col = self.ts_analysis['tss'].order_by[0]
            if self.grouped_by == ['__default']:
                series_idxs = data['data'][f'__mdb_original_{self.ts_analysis["tss"].order_by[0]}']
                series_data = data['data'][self.target].values
                series_oby = data['data'][oby_col].values
            else:
                target_idx = data['data'].columns.tolist().index(self.target)
                oby_idx = data['data'].columns.tolist().index(oby_col)
                series_idxs, series_data = get_group_matches(data, group)
                series_oby = series_data[:, oby_idx]
                series_data = series_data[:, target_idx]

            if series_data.size > self.ts_analysis['tss'].window:
                series = pd.Series(series_data.squeeze(), index=series_idxs)
                series = series.sort_index(ascending=True)
                series = series.reset_index(drop=True)
                series = series.loc[~pd.isnull(series.values)]  # remove NaN  # @TODO: benchmark imputation vs this?

                if self.model_path == 'fbprophet.Prophet':
                    series = self._transform_index_to_datetime(series, series_oby, options['freq'])

                series = series.astype(float)

                # if data is huge, filter out old records for quicker fitting
                if self.auto_size:
                    cutoff = min(len(series), max(500, options['sp'] * self.cutoff_factor))
                    series = series.iloc[-cutoff:]
                try:
                    self.models[group].fit(series, fh=self.fh)
                except Exception:
                    self.models[group] = model_class()  # with default options (i.e. no seasonality, among others)
                    self.models[group].fit(series, fh=self.fh)

    def partial_fit(self, train_data: EncodedDs, dev_data: EncodedDs) -> None:
        """
        Note: sktime asks for "specification of the time points for which forecasts are requested", and this mixer complies by assuming forecasts will start immediately after the last observed value.

        Because of this, `ProblemDefinition.fit_on_all` is set to True so that `partial_fit` uses both `dev` and `test` splits to fit the models.

        Due to how lightwood implements the `update` procedure, expected inputs for this method are:

        :param dev_data: original `test` split (used to validate and select model if ensemble is `BestOf`).
        :param train_data: concatenated original `train` and `dev` splits.
        """  # noqa
        self.hyperparam_search = False
        self.fit(dev_data, train_data)
        self.prepared = True

    def __call__(self, ds: Union[EncodedDs, ConcatedEncodedDs],
                 args: PredictionArguments = PredictionArguments()) -> pd.DataFrame:
        """
        Calls the mixer to emit forecasts.
        
        If there are groups that were not observed at training, a default forecaster (trained on all available data) is used, warning the user that performance might not be optimal.
        
        Latest data point in `train_data` passed to `fit()` determines the starting point for predictions. Relative offsets can be provided to forecast through the `args.forecast_offset` argument.
        """  # noqa
        if args.predict_proba:
            log.warning('This mixer does not output probability estimates')

        forecast_offset = args.forecast_offset
        if '__mdb_forecast_offset' in ds.data_frame.columns:
            if ds.data_frame['__mdb_forecast_offset'].nunique() == 1:
                forecast_offset = int(ds.data_frame['__mdb_forecast_offset'].unique()[0])

        length = sum(ds.encoded_ds_lenghts) if isinstance(ds, ConcatedEncodedDs) else len(ds)
        ydf = pd.DataFrame(0,  # zero-filled
                           index=np.arange(length),
                           columns=['prediction'],
                           dtype=object)

        data = {'data': ds.data_frame[self.target],
                'group_info': {gcol: ds.data_frame[gcol].tolist()
                               for gcol in self.grouped_by} if self.ts_analysis['tss'].group_by
                else {'': ['__default' for _ in range(length)]}}

        pending_idxs = set(range(length))
        all_group_combinations = list(product(*[set(x) for x in data['group_info'].values()]))
        for group in all_group_combinations:
            group = tuple(group)
            group = '__default' if group[0] == '__default' else group
            series_idxs, series_data = get_group_matches(data, group)

            if series_data.size > 0:
                series_idxs = sorted(series_idxs)
                if self.models.get(group, False) and self.models[group].is_fitted:
                    forecaster = self.models[group]
                    series = pd.Series(series_data.squeeze(), index=series_idxs)
                    ydf = self._call_groupmodel(ydf, forecaster, series, offset=forecast_offset)
                else:
                    log.warning(f"Applying naive forecaster for novel group {group}. Performance might not be optimal.")
                    ydf = self._call_default(ydf, series_data, series_idxs)
                pending_idxs -= set(series_idxs)

        # apply default model in all remaining novel-group rows
        if len(pending_idxs) > 0:
            series = data['data'][list(pending_idxs)].squeeze()
            ydf = self._call_default(ydf, series, list(pending_idxs))

        return ydf[['prediction']]

    def _call_groupmodel(self,
                         ydf: pd.DataFrame,
                         model: BaseForecaster,
                         series: pd.Series,
                         offset: int = 0):
        """
        Inner method that calls a `sktime.BaseForecaster`.

        :param offset: indicates relative offset to the latest data point seen during model training. Cannot be less than the number of training data points + the amount of diffences applied internally by the model.
        """  # noqa
        original_index = series.index
        series = series.reset_index(drop=True)

        if isinstance(model, TransformedTargetForecaster):
            submodel = model.steps_[-1][-1]
        else:
            submodel = model

        if hasattr(submodel, '_cutoff') and hasattr(submodel, 'd'):
            model_d = 0 if submodel.d is None else submodel.d
            min_offset = -submodel._cutoff + model_d + 1
        else:
            min_offset = -np.inf

        start = max(1 + offset, min_offset)
        end = 1 + series.shape[0] + offset + self.horizon
        all_preds = model.predict(np.arange(start, end)).tolist()

        for idx, _ in enumerate(series.iteritems()):
            start_idx = 0 if max(1 + idx + offset, min_offset) < 0 else idx
            end_idx = start_idx + self.horizon
            ydf['prediction'].iloc[original_index[idx]] = all_preds[start_idx:end_idx]
        return ydf

    def _call_default(self, ydf, data, idxs):
        # last value from each window equals shifted target (by 1)  # noqa
        series = np.array([0] + list(data.flatten())[:-1])
        all_preds = [[value for _ in range(self.horizon)] for value in series]
        ydf['prediction'].iloc[idxs] = all_preds
        return ydf

    def _get_best_model(self, trial, train_data, test_data):
        """
        Helper function for Optuna hyperparameter optimization.
        For now, it uses dev data split to find the best model out of the list specified in self.possible_models.
        """

        self.hyperparam_dict = {
            'class': trial.suggest_categorical('class', self.possible_models)
        }
        log.info(f'Starting trial with hyperparameters: {self.hyperparam_dict}')
        try:
            self._fit(train_data)
            y_true = test_data.data_frame[self.target].values[:self.horizon]
            y_pred = self(test_data)['prediction'].iloc[0][:len(y_true)]
            error = self.trial_error_fn(y_true, y_pred)
        except Exception as e:
            log.debug(e)
            error = np.inf

        log.info(f'Trial got error: {error}')
        return error

    def _add_forecaster_kwarg(self, forecaster: BaseForecaster, kwargs: dict, key: str, value):
        """
        Adds arguments to the `kwargs` dictionary if the key-value pair is valid for the `forecaster` class signature.
        """
        if key in [p.name for p in inspect.signature(forecaster).parameters.values()]:
            kwargs[key] = value

        return kwargs

    def _transform_index_to_datetime(self, series, series_oby, freq):
        series_oby = np.array([np.array(lst) for lst in series_oby])
        start = datetime.utcfromtimestamp(np.min(series_oby[series_oby != np.min(series_oby)]))
        series.index = pd.date_range(start=start, freq=freq, normalize=False, periods=series.shape[0])
        return series

    def _get_freq(self, delta):
        labels = ['S', 'T', 'H', 'D', 'W', 'M', 'Q', 'Y']
        secs = [1,
                60,
                60 * 60,
                60 * 60 * 24,
                60 * 60 * 24 * 7,
                60 * 60 * 24 * 7 * 4,
                60 * 60 * 24 * 7 * 4 * 3,
                60 * 60 * 24 * 7 * 4 * 12]
        min_diff = np.argmin(np.abs(np.array(secs) - delta))
        return labels[min_diff]



from lightwood.helpers.accuracy import to_binary, f1_score, recall_score, precision_score, r2_score
from lightwood.helpers.device import is_cuda_compatible, get_devices
from lightwood.helpers.general import get_group_matches, mase, is_none, evaluate_accuracy, evaluate_num_array_accuracy,\
    evaluate_array_accuracy, evaluate_multilabel_accuracy, evaluate_regression_accuracy, evaluate_cat_array_accuracy, \
    bounded_evaluate_array_accuracy
from lightwood.helpers.io import read_from_path_or_url
from lightwood.helpers.parallelism import get_nr_procs, mut_method_call, run_mut_method
from lightwood.helpers.numeric import is_nan_numeric, filter_nan_and_none
from lightwood.helpers.seed import seed
from lightwood.helpers.text import tokenize_text, analyze_sentences, decontracted, contains_alnum,\
    get_identifier_description, get_identifier_description_mp, get_pct_auto_increment, extract_digits, isascii,\
    hashtext, splitRecursive, cast_string_to_python_type, gen_chars, clean_float, word_tokenize, get_language_dist
from lightwood.helpers.torch import average_vectors, concat_vectors_and_pad, LightwoodAutocast


__all__ = ['to_binary', 'f1_score', 'recall_score', 'precision_score', 'r2_score', 'is_cuda_compatible', 'get_devices',
           'get_group_matches', 'mase', 'is_none', 'evaluate_accuracy', 'evaluate_num_array_accuracy',
           'evaluate_array_accuracy', 'evaluate_cat_array_accuracy', 'bounded_evaluate_array_accuracy',
           'evaluate_multilabel_accuracy', 'evaluate_regression_accuracy', 'read_from_path_or_url', 'get_nr_procs',
           'mut_method_call', 'run_mut_method', 'tokenize_text', 'analyze_sentences', 'decontracted', 'contains_alnum',
           'get_identifier_description', 'get_identifier_description_mp', 'get_pct_auto_increment',
           'extract_digits', 'isascii',
           'hashtext', 'splitRecursive', 'cast_string_to_python_type', 'gen_chars', 'clean_float', 'word_tokenize',
           'get_language_dist', 'average_vectors', 'concat_vectors_and_pad', 'LightwoodAutocast', 'is_nan_numeric', 'filter_nan_and_none', 'seed']  # noqa



import unittest

from torch import Tensor
from lightwood.encoder.image import Img2VecEncoder
import os


class TestImg2VecEncoder(unittest.TestCase):
    def test_encode(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))

        image_path = os.path.join(
            dir_path, 'test_image.jpg'
        )

        enc = Img2VecEncoder()
        enc.prepare([], [])
        encoded_images_tensor = enc.encode(images=[image_path])

        self.assertTrue(isinstance(encoded_images_tensor, Tensor))

        # We expect the first dimension to equal the number of images
        self.assertEqual(encoded_images_tensor.size(0), 1)

        # We expect the second dimesion to be 512
        # NOTE: this will break when it will possible to choose different
        # encoding models.
        self.assertEqual(encoded_images_tensor.size(1), 512)



from typing import List, Dict
from itertools import product

import numpy as np
import pandas as pd

from lightwood.helpers.log import log
from lightwood.api.dtype import dtype
from lightwood.api.types import TimeseriesSettings


def splitter(
        data: pd.DataFrame,
        tss: TimeseriesSettings,
        dtype_dict: Dict[str, str],
        seed: int,
        pct_train: float,
        pct_dev: float,
        pct_test: float,
        target: str
) -> Dict[str, pd.DataFrame]:
    """
    Splits data into training, dev and testing datasets. 
    
    The proportion of data for each split must be specified (JSON-AI sets defaults to 80/10/10). First, rows in the dataset are shuffled randomly. Then a simple split is done. If a target value is provided and is of data type categorical/binary, then the splits will be stratified to maintain the representative populations of each class.

    :param data: Input dataset to be split
    :param tss: time-series specific details for splitting
    :param dtype_dict: Dictionary with the data type of all columns
    :param seed: Random state for pandas data-frame shuffling
    :param pct_train: training fraction of data; must be less than 1
    :param pct_dev: dev fraction of data; must be less than 1
    :param pct_test: testing fraction of data; must be less than 1
    :param target: Name of the target column; if specified, data will be stratified on this column

    :returns: A dictionary containing the keys train, test and dev with their respective data frames, as well as the "stratified_on" key indicating which columns the data was stratified on (None if it wasn't stratified on anything)
    """  # noqa
    pct_sum = pct_train + pct_dev + pct_test
    if not (np.isclose(pct_sum, 1, atol=0.001) and np.less(pct_sum, 1 + 1e-5)):
        raise Exception(f'The train, dev and test percentage of the data needs to sum up to 1 (got {pct_sum})')

    # Shuffle the data
    np.random.seed(seed)
    if not tss.is_timeseries:
        data = data.sample(frac=1, random_state=seed).reset_index(drop=True)

    # Check if stratification should be done
    stratify_on = []
    if target is not None:
        if dtype_dict[target] in (dtype.categorical, dtype.binary) and not tss.is_timeseries:
            stratify_on = [target]
        if tss.is_timeseries and isinstance(tss.group_by, list):
            stratify_on = tss.group_by

    # Split the data
    if stratify_on:
        reshuffle = not tss.is_timeseries
        train, dev, test = stratify(data, pct_train, pct_dev, pct_test, stratify_on, seed, reshuffle)
    else:
        train, dev, test = simple_split(data, pct_train, pct_dev, pct_test)

    return {"train": train, "test": test, "dev": dev, "stratified_on": stratify_on}


def simple_split(data: pd.DataFrame,
                 pct_train: float,
                 pct_dev: float,
                 pct_test: float) -> List[pd.DataFrame]:
    """
    Simple split method to separate data into training, dev and testing datasets.

    :param data: Input dataset to be split
    :param pct_train: training fraction of data; must be less than 1
    :param pct_dev: dev fraction of data; must be less than 1
    :param pct_test: testing fraction of data; must be less than 1

    :returns Train, dev, and test dataframes
    """
    train_cutoff = round(data.shape[0] * pct_train)
    dev_cutoff = round(data.shape[0] * pct_dev) + train_cutoff
    test_cutoff = round(data.shape[0] * pct_test) + dev_cutoff

    train = data[:train_cutoff]
    dev = data[train_cutoff:dev_cutoff]
    test = data[dev_cutoff:test_cutoff]

    return [train, dev, test]


def stratify(data: pd.DataFrame,
             pct_train: float,
             pct_dev: float,
             pct_test: float,
             stratify_on: List[str],
             seed: int,
             reshuffle: bool) -> List[pd.DataFrame]:
    """
    Stratified data splitter.

    The `stratify_on` columns yield a cartesian product by which every different subset will be stratified
    independently from the others, and recombined at the end in fractions specified by `pcts`.

    For grouped time series tasks, stratification is done based on the group-by columns.

    :param data: dataframe with data to be split
    :param pct_train: fraction of data to use for training split
    :param pct_dev: fraction of data to use for dev split (used internally by mixers)
    :param pct_test: fraction of data to use for test split (used post-training for analysis)
    :param stratify_on: Columns to consider when stratifying
    :param seed: Random state for pandas data-frame shuffling
    :param reshuffle: specify if reshuffling should be done post-split

    :returns Stratified train, dev, test dataframes
    """  # noqa

    train_st = pd.DataFrame(columns=data.columns)
    dev_st = pd.DataFrame(columns=data.columns)
    test_st = pd.DataFrame(columns=data.columns)

    all_group_combinations = list(product(*[data[col].unique() for col in stratify_on]))
    for group in all_group_combinations:
        df = data
        for idx, col in enumerate(stratify_on):
            df = df[df[col] == group[idx]]

        train_cutoff = round(df.shape[0] * pct_train)
        dev_cutoff = round(df.shape[0] * pct_dev) + train_cutoff
        test_cutoff = round(df.shape[0] * pct_test) + dev_cutoff

        train_st = train_st.append(df[:train_cutoff])
        dev_st = dev_st.append(df[train_cutoff:dev_cutoff])
        test_st = test_st.append(df[dev_cutoff:test_cutoff])

    if reshuffle:
        train_st, dev_st, test_st = [df.sample(frac=1, random_state=seed).reset_index(drop=True)
                                     for df in [train_st, dev_st, test_st]]

    # check that stratified lengths conform to expected percentages
    if not np.isclose(len(train_st) / len(data), pct_train, atol=0.01) or \
            not np.isclose(len(dev_st) / len(data), pct_dev, atol=0.01) or \
            not np.isclose(len(test_st) / len(data), pct_test, atol=0.01):
        log.info("Could not stratify; reverting to simple split")
        train_st, dev_st, test_st = simple_split(data, pct_train, pct_dev, pct_test)

    return [train_st, dev_st, test_st]



from copy import deepcopy
from itertools import product
from typing import Dict, Tuple
from types import SimpleNamespace

import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from lightwood.api.dtype import dtype
from lightwood.api.types import PredictionArguments
from lightwood.helpers.ts import add_tn_conf_bounds

from lightwood.data import EncodedDs
from lightwood.analysis.base import BaseAnalysisBlock
from lightwood.analysis.nc.norm import Normalizer
from lightwood.analysis.nc.icp import IcpRegressor, IcpClassifier
from lightwood.analysis.nc.base import CachedRegressorAdapter, CachedClassifierAdapter
from lightwood.analysis.nc.nc import BoostedAbsErrorErrFunc, RegressorNc, ClassifierNc, MarginErrFunc
from lightwood.analysis.nc.util import clean_df, set_conf_range, get_numeric_conf_range, \
    get_categorical_conf, get_anomalies


class ICP(BaseAnalysisBlock):
    """ Confidence estimation block, uses inductive conformal predictors (ICPs) for model agnosticity """

    def __init__(self,
                 fixed_significance: float,
                 positive_domain: bool,
                 confidence_normalizer: bool
                 ):
        super().__init__()
        self.fixed_significance = fixed_significance
        self.positive_domain = positive_domain
        self.confidence_normalizer = confidence_normalizer
        self.validation_size = 100  # determines size of nonconformity score arrays (has sizable impact in runtime)

    def analyze(self, info: Dict[str, object], **kwargs) -> Dict[str, object]:
        ns = SimpleNamespace(**kwargs)

        data_type = ns.dtype_dict[ns.target]
        output = {'icp': {'__mdb_active': False}}

        fit_params = {'horizon': ns.tss.horizon or 0, 'columns_to_ignore': []}
        fit_params['columns_to_ignore'].extend([f'timestep_{i}' for i in range(1, fit_params['horizon'])])

        if ns.is_classification:
            if ns.predictor.supports_proba:
                all_cat_cols = [col for col in ns.normal_predictions.columns
                                if '__mdb_proba' in col and '__mdb_unknown_cat' not in col]
                all_classes = np.array([col.replace('__mdb_proba_', '') for col in all_cat_cols])
            else:
                class_keys = sorted(ns.encoded_val_data.encoders[ns.target].rev_map.keys())
                all_classes = np.array([ns.encoded_val_data.encoders[ns.target].rev_map[idx] for idx in class_keys])

            if data_type != dtype.tags:
                enc = OneHotEncoder(sparse=False, handle_unknown='ignore')
                enc.fit(all_classes.reshape(-1, 1))
                output['label_encoders'] = enc  # needed to repr cat labels inside nonconformist
            else:
                output['label_encoders'] = None

            adapter = CachedClassifierAdapter
            nc_function = MarginErrFunc()
            nc_class = ClassifierNc
            icp_class = IcpClassifier

        else:
            adapter = CachedRegressorAdapter
            nc_function = BoostedAbsErrorErrFunc()
            nc_class = RegressorNc
            icp_class = IcpRegressor

        result_df = pd.DataFrame()

        if ns.is_numerical or (ns.is_classification and data_type != dtype.tags):
            model = adapter(ns.predictor)

            norm_params = {'target': ns.target, 'dtype_dict': ns.dtype_dict, 'predictor': ns.predictor,
                           'encoders': ns.encoded_val_data.encoders, 'is_multi_ts': ns.is_multi_ts, 'stop_after': 1e2}
            if self.confidence_normalizer:
                normalizer = Normalizer(fit_params=norm_params)
                normalizer.fit(ns.train_data)
                normalizer.prediction_cache = normalizer(ns.encoded_val_data, args=PredictionArguments())
            else:
                normalizer = None

            # instance the ICP
            nc = nc_class(model, nc_function, normalizer=normalizer)
            icp = icp_class(nc, cal_size=self.validation_size)

            output['icp']['__default'] = icp

            # setup prediction cache to avoid additional .predict() calls
            if ns.is_classification:
                if ns.predictor.supports_proba:
                    icp.nc_function.model.prediction_cache = ns.normal_predictions[all_cat_cols].values
                else:
                    if ns.is_multi_ts:
                        icp.nc_function.model.prediction_cache = np.array(
                            [p[0] for p in ns.normal_predictions['prediction']])
                        preds = icp.nc_function.model.prediction_cache
                    else:
                        preds = ns.normal_predictions['prediction']
                    predicted_classes = pd.get_dummies(preds).values  # inflate to one-hot enc
                    icp.nc_function.model.prediction_cache = predicted_classes

            elif ns.is_multi_ts:
                # we fit ICPs for time series confidence bounds only at t+1 forecast
                icp.nc_function.model.prediction_cache = np.array([p[0] for p in ns.normal_predictions['prediction']])
            else:
                icp.nc_function.model.prediction_cache = np.array(ns.normal_predictions['prediction'])

            if not ns.is_classification:
                output['df_target_stddev'] = {'__default': ns.stats_info.df_target_stddev}

            # fit additional ICPs in time series tasks with grouped columns
            if ns.tss.is_timeseries and ns.tss.group_by:

                # create an ICP for each possible group
                group_info = ns.data[ns.tss.group_by].to_dict('list')
                all_group_combinations = list(product(*[set(x) for x in group_info.values()]))
                output['icp']['__mdb_groups'] = all_group_combinations
                output['icp']['__mdb_group_keys'] = [x for x in group_info.keys()]

                for combination in all_group_combinations:
                    output['icp'][tuple(combination)] = deepcopy(icp)

            # calibrate ICP
            icp_df = deepcopy(ns.data)
            icp_df, y = clean_df(icp_df, ns.target, ns.is_classification, output.get('label_encoders', None))
            output['icp']['__default'].index = icp_df.columns
            output['icp']['__default'].calibrate(icp_df.values, y)

            # get confidence estimation for validation dataset
            conf, ranges = set_conf_range(
                icp_df, icp, ns.dtype_dict[ns.target],
                output, positive_domain=self.positive_domain, significance=self.fixed_significance)
            if not ns.is_classification:
                result_df = pd.DataFrame(index=ns.data.index, columns=['confidence', 'lower', 'upper'], dtype=float)
                result_df.loc[icp_df.index, 'lower'] = ranges[:, 0]
                result_df.loc[icp_df.index, 'upper'] = ranges[:, 1]
            else:
                result_df = pd.DataFrame(index=ns.data.index, columns=['confidence'], dtype=float)

            result_df.loc[icp_df.index, 'confidence'] = conf

            # calibrate additional grouped ICPs
            if ns.tss.is_timeseries and ns.tss.group_by:
                icps = output['icp']
                group_keys = icps['__mdb_group_keys']

                # add all predictions to DF
                icps_df = deepcopy(ns.data)
                if ns.is_multi_ts:
                    icps_df[f'__predicted_{ns.target}'] = [p[0] for p in ns.normal_predictions['prediction']]
                else:
                    icps_df[f'__predicted_{ns.target}'] = ns.normal_predictions['prediction']

                for group in icps['__mdb_groups']:
                    icp_df = icps_df

                    # filter irrelevant rows for each group combination
                    icp_df['__mdb_norm_index'] = np.arange(len(icp_df))
                    for key, val in zip(group_keys, group):
                        icp_df = icp_df[icp_df[key] == val]

                    if icps[tuple(group)].nc_function.normalizer is not None:
                        group_normalizer = icps[tuple(group)].nc_function.normalizer
                        norm_input_df = ns.encoded_val_data.data_frame.iloc[icp_df.pop('__mdb_norm_index')]
                        norm_input = EncodedDs(ns.encoded_val_data.encoders, norm_input_df, ns.target)
                        norm_cache = group_normalizer(norm_input, args=PredictionArguments())
                        icp_df[f'__norm_{ns.target}'] = norm_cache

                    # save relevant predictions in the caches, then calibrate the ICP
                    pred_cache = icp_df.pop(f'__predicted_{ns.target}').values
                    icps[tuple(group)].nc_function.model.prediction_cache = pred_cache
                    icp_df, y = clean_df(icp_df, ns.target, ns.is_classification, output.get('label_encoders', None))
                    if icps[tuple(group)].nc_function.normalizer is not None:
                        icps[tuple(group)].nc_function.normalizer.prediction_cache = icp_df.pop(
                            f'__norm_{ns.target}').values

                    icps[tuple(group)].index = icp_df.columns  # important at inference time
                    icps[tuple(group)].calibrate(icp_df.values, y)

                    # save training std() for bounds width selection
                    if not ns.is_classification:
                        icp_train_df = ns.data
                        for key, val in zip(group_keys, group):
                            icp_train_df = icp_train_df[icp_train_df[key] == val]
                        y_train = icp_train_df[ns.target].values
                        output['df_target_stddev'][tuple(group)] = y_train.std()

                    # get bounds for relevant rows in validation dataset
                    conf, group_ranges = set_conf_range(
                        icp_df, icps[tuple(group)],
                        ns.dtype_dict[ns.target],
                        output, group=tuple(group),
                        positive_domain=self.positive_domain, significance=self.fixed_significance)
                    # save group bounds
                    if not ns.is_classification:
                        result_df.loc[icp_df.index, 'lower'] = group_ranges[:, 0]
                        result_df.loc[icp_df.index, 'upper'] = group_ranges[:, 1]

                    result_df.loc[icp_df.index, 'confidence'] = conf

            # consolidate all groups here
            output['icp']['__mdb_active'] = True

        output['result_df'] = result_df

        info = {**info, **output}
        return info

    def explain(self, row_insights: pd.DataFrame, global_insights: Dict[str, object],
                **kwargs) -> Tuple[pd.DataFrame, Dict[str, object]]:
        ns = SimpleNamespace(**kwargs)

        if ns.analysis['icp']['__mdb_active']:
            icp_X = deepcopy(ns.data)

            # replace observed data w/predictions
            preds = ns.predictions['prediction']
            if ns.tss.is_timeseries and ns.tss.horizon > 1:
                preds = [p[0] for p in preds]

                for col in [f'timestep_{i}' for i in range(1, ns.tss.horizon)]:
                    if col in icp_X.columns:
                        icp_X.pop(col)  # erase ignorable columns

            icp_X[ns.target_name] = preds

            is_categorical = ns.target_dtype in (dtype.binary, dtype.categorical, dtype.cat_array, dtype.cat_tsarray)
            is_numerical = ns.target_dtype in (dtype.integer, dtype.float,
                                               dtype.quantity, dtype.num_array, dtype.num_tsarray)
            is_anomaly_task = is_numerical and ns.tss.is_timeseries and ns.anomaly_detection

            if (is_numerical or is_categorical) and ns.analysis['icp'].get('__mdb_active', False):
                base_icp = ns.analysis['icp']['__default']
                # reorder DF index
                index = base_icp.index.values
                index = np.append(index, ns.target_name) if ns.target_name not in index else index
                icp_X = icp_X.reindex(columns=index)  # important, else bounds can be invalid

                # only one normalizer, even if it's a grouped time series task
                normalizer = base_icp.nc_function.normalizer
                if normalizer:
                    normalizer.prediction_cache = normalizer(ns.encoded_data, args=PredictionArguments)
                    icp_X['__mdb_selfaware_scores'] = normalizer.prediction_cache

                # get ICP predictions
                result_cols = ['lower', 'upper', 'significance'] if is_numerical else ['significance']
                result = pd.DataFrame(index=icp_X.index, columns=result_cols)

                # base ICP
                X = deepcopy(icp_X)
                # Calling `values` multiple times increased runtime of this function; referenced var is faster
                icp_values = X.values

                # get all possible ranges
                if ns.tss.is_timeseries and ns.tss.horizon > 1 and is_numerical:

                    # bounds in time series are only given for the first forecast
                    base_icp.nc_function.model.prediction_cache = preds
                    all_confs = base_icp.predict(icp_values)

                elif is_numerical:
                    base_icp.nc_function.model.prediction_cache = preds
                    all_confs = base_icp.predict(icp_values)

                # categorical
                else:
                    predicted_proba = True if any(['__mdb_proba' in col for col in ns.predictions.columns]) else False
                    if predicted_proba:
                        all_cat_cols = [col for col in ns.predictions.columns
                                        if '__mdb_proba' in col and '__mdb_unknown_cat' not in col]
                        class_dists = ns.predictions[all_cat_cols].values
                        for icol, cat_col in enumerate(all_cat_cols):
                            row_insights.loc[X.index, cat_col] = class_dists[:, icol]
                    else:
                        class_dists = pd.get_dummies(preds).values

                    base_icp.nc_function.model.prediction_cache = class_dists

                    all_ranges = np.array([base_icp.predict(icp_values)])
                    all_confs = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)

                # convert (B, 2, 99) into (B, 2) given width or error rate constraints
                if is_numerical:
                    significances, confs = get_numeric_conf_range(all_confs,
                                                                  df_target_stddev=ns.analysis['df_target_stddev'],
                                                                  positive_domain=self.positive_domain,
                                                                  fixed_conf=ns.pred_args.fixed_confidence)
                    result.loc[X.index, 'lower'] = confs[:, 0]
                    result.loc[X.index, 'upper'] = confs[:, 1]
                else:
                    significances = get_categorical_conf(all_confs.squeeze())

                result.loc[X.index, 'significance'] = significances

                # grouped time series, we replace bounds in rows that have a trained ICP
                if ns.analysis['icp'].get('__mdb_groups', False):
                    icps = ns.analysis['icp']
                    group_keys = icps['__mdb_group_keys']

                    for group in icps['__mdb_groups']:
                        icp = icps[tuple(group)]

                        # check ICP has calibration scores
                        if icp.cal_scores[0].shape[0] > 0:

                            # filter rows by group
                            X = deepcopy(icp_X)
                            for key, val in zip(group_keys, group):
                                X = X[X[key] == val]

                            if X.size > 0:
                                # set ICP caches
                                icp.nc_function.model.prediction_cache = X.pop(ns.target_name).values
                                if icp.nc_function.normalizer:
                                    icp.nc_function.normalizer.prediction_cache = X.pop('__mdb_selfaware_scores').values

                                # predict and get confidence level given width or error rate constraints
                                if is_numerical:
                                    all_confs = icp.predict(X.values)
                                    fixed_conf = ns.pred_args.fixed_confidence
                                    significances, confs = get_numeric_conf_range(
                                        all_confs,
                                        df_target_stddev=ns.analysis['df_target_stddev'],
                                        positive_domain=self.positive_domain,
                                        group=tuple(group),
                                        fixed_conf=fixed_conf
                                    )

                                    # only replace where grouped ICP is more informative (i.e. tighter)
                                    if fixed_conf is None:
                                        default_widths = result.loc[X.index, 'upper'] - result.loc[X.index, 'lower']
                                        grouped_widths = np.subtract(confs[:, 1], confs[:, 0])
                                        insert_index = (default_widths > grouped_widths)[lambda x: x.isin([True])].index
                                        conf_index = (default_widths.reset_index(drop=True) >
                                                      grouped_widths)[lambda x: x.isin([True])].index

                                        result.loc[insert_index, 'lower'] = confs[conf_index, 0]
                                        result.loc[insert_index, 'upper'] = confs[conf_index, 1]
                                        result.loc[insert_index, 'significance'] = significances[conf_index]

                                else:
                                    all_ranges = np.array([icp.predict(X.values)])
                                    all_confs = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)
                                    significances = get_categorical_conf(all_confs)
                                    result.loc[X.index, 'significance'] = significances

                row_insights['confidence'] = result['significance'].astype(float).tolist()

                if is_numerical:
                    row_insights['lower'] = result['lower'].astype(float)
                    row_insights['upper'] = result['upper'].astype(float)

                # anomaly detection
                if is_anomaly_task:
                    anomalies = get_anomalies(row_insights,
                                              ns.data[ns.target_name],
                                              cooldown=ns.pred_args.anomaly_cooldown)
                    row_insights['anomaly'] = anomalies

            if ns.tss.is_timeseries and ns.tss.horizon > 1 and is_numerical:
                row_insights = add_tn_conf_bounds(row_insights, ns.tss)

            # clip bounds if necessary
            if is_numerical:
                lower_limit = 0.0 if ns.positive_domain else -pow(2, 62)
                upper_limit = pow(2, 62)
                if not (ns.tss.is_timeseries and ns.tss.horizon > 1):
                    row_insights['upper'] = row_insights['upper'].clip(lower_limit, upper_limit)
                    row_insights['lower'] = row_insights['lower'].clip(lower_limit, upper_limit)
                else:
                    row_insights['upper'] = [np.array(row).clip(lower_limit, upper_limit).tolist()
                                             for row in row_insights['upper']]
                    row_insights['lower'] = [np.array(row).clip(lower_limit, upper_limit).tolist()
                                             for row in row_insights['lower']]

            # Make sure the target and real values are of an appropriate type
            if ns.tss.is_timeseries and ns.tss.horizon > 1:
                # Array output that are not of type <array> originally are odd and I'm not sure how to handle them
                # Or if they even need handling yet
                pass
            elif ns.target_dtype in (dtype.integer):
                row_insights['prediction'] = row_insights['prediction'].astype(int)
                row_insights['upper'] = row_insights['upper'].astype(int)
                row_insights['lower'] = row_insights['lower'].astype(int)

            elif ns.target_dtype in (dtype.float, dtype.quantity):
                row_insights['prediction'] = row_insights['prediction'].astype(float)
                row_insights['upper'] = row_insights['upper'].astype(float)
                row_insights['lower'] = row_insights['lower'].astype(float)

            elif ns.target_dtype in (dtype.short_text, dtype.rich_text, dtype.binary, dtype.categorical):
                row_insights['prediction'] = row_insights['prediction'].astype(str)

        return row_insights, global_insights



from lightwood.api.predictor import PredictorInterface
import time
import pandas as pd


def train_and_check_time_aim(predictor: PredictorInterface, train_df: pd.DataFrame):
    """
    Trains the predictor with the desired data *and* checks that the time aim we set for it was respected.
    """
    time_aim_expected = predictor.problem_definition.time_aim
    start = time.time()
    predictor.learn(train_df)
    time_aim_actual = (time.time() - start)
    # Current margin of error for time aim is 2x, but we should aim to lower this
    if time_aim_expected is not None:
        if((time_aim_expected * 2.5) < time_aim_actual):
            error = f'time_aim is set to {time_aim_expected} seconds, however learning took {time_aim_actual}'
            raise ValueError(error)
